import unittest

import numpy as np
from scipy.stats import entropy

from epipack.stochastic_epi_models import (
            StochasticEpiModel,
            StochasticSIModel,
            StochasticSIRModel,
            StochasticSISModel,
        )

class StochasticEpiTest(unittest.TestCase):

    def test_compartments(self):
        epi = StochasticEpiModel(list("SEIR"),10)
        assert(all([ i == epi.get_compartment_id(C) for i, C in enumerate("SEIR") ]))
        assert(epi.get_compartment_id("E") == 1)
        assert(epi.get_compartment(1) == "E")

    def test_mean_contact_number(self):
        self.assertRaises(TypeError,StochasticEpiModel,list("SEIR"),10,well_mixed_mean_contact_number=0.5)

    def test_errors(self):
        epi = StochasticEpiModel(list("SEIR"),10)
        self.assertRaises(ValueError,epi.set_link_transmission_processes,[("S","I",1.0,"I","I")])

        # error: first compartment has to be infecting compartment
        self.assertRaises(ValueError,epi.set_conditional_link_transmission_processes,
                    { ("S", "I", "->", "I", "I") :
                        [
                            ("I", "R", "->", "I", "I"),
                        ]
                    }
                )

        # error: 4-length tuples are invalid processes
        self.assertRaises(ValueError,epi.set_conditional_link_transmission_processes,
                    { ("S", "I", "I", "I") :
                        [
                            ("I", "R", "->", "I", "I"),
                        ]
                    }
                )

        # error triggered events can only be of type transmission
        self.assertRaises(ValueError,epi.set_conditional_link_transmission_processes,
                    { ("I", "S", "->", "I", "I") :
                        [
                            ("R", "->", "I"),
                        ]
                    }
                )

        # error: target compartment of triggering event has to be first source compartment
        # of triggered event
        self.assertRaises(ValueError,epi.set_conditional_link_transmission_processes,
                    { ("I", "S", "->", "I", "I") :
                        [
                            ("R", "I", "->", "I", "I"),
                        ]
                    }
                )
            
        # error: "The source (infecting) compartment", coupling0, "must be equal to the affected (target) compartment of",

        #"the triggering event", triggering_event, " but is not."
        self.assertRaises(ValueError,epi.set_conditional_link_transmission_processes,
                    { ("I", "S", "->", "I", "I") :
                        [
                            ("R", "S", "->", "R", "I"),
                        ]
                    }
                )

        #probabilities sum to value large than one
        self.assertRaises(ValueError,epi.set_conditional_link_transmission_processes,
                    { ("I", "S", "->", "I", "I") :
                        [
                            ("I", "R", 0.6, "I", "E"),
                            ("I", "R", 0.6, "I", "S"),
                        ]
                    }
                )

        # error: double entry
        self.assertRaises(ValueError,epi.set_random_initial_conditions,
                    [ ("S",2),("S",1) ]
                    )

        # error: sum of initial conditions
        self.assertRaises(ValueError,epi.set_random_initial_conditions,
                    [ ("S",5),("I",1) ]
                    )

        # error: not the right amount of node statuses
        self.assertRaises(ValueError,epi.set_node_statuses,[0,1])

        self.assertRaises(ValueError,epi.simulate,10,sampling_callback=lambda x: x)

    def test_random_initial_conditions(self):
        epi = StochasticEpiModel(list("SEIR"),10)
        epi.set_node_transition_processes([("I",1,"R")])
        epi.set_random_initial_conditions({
                "S": 4,
                "I": 6,
                "R": 0,
            })

        assert(np.count_nonzero(epi.node_status==0) == 4)
        assert(np.count_nonzero(epi.node_status==2) == 6)
        assert(np.count_nonzero(epi.node_status==3) == 0)



    def test_rate_evaluation(self):
        S, S0,S1, I, A, B, C0, C1, D, E, F = "S S0 S1 I A B C0 C1 D E F".split(" ")

        rateA = 3.0
        rateB = 2.0
        rateE = 1.0
        rateF = 5.0
        rateC = 10.0


        N = 6
        edges = [ (0, i, 1.0) for i in range(1,N) ]
        edges.extend([(N+1, 0, 1.0),(N,0,1.0),(N+1,1,1.0), (N+1,2,1.0)])

        model = StochasticEpiModel([S,S0,S1,I,A,B,C0,C1,D, E, F], N+2, edges)

        model.set_node_transition_processes([
                (S0, rateC, E),
                (I, rateF, F),
            ])

        model.set_link_transmission_processes([
                (S0, I, rateA, S0, A),
                (S0, I, rateB, S0, B),
                (S1, I, rateE, S1, E),
            ])

        statuses = np.zeros(N+2,dtype=int)
        statuses[0] = model.get_compartment_id(I)
        statuses[N] = model.get_compartment_id(S0)
        statuses[N+1] = model.get_compartment_id(S1)

        model.set_node_statuses(statuses)

        expected_rate = rateA + rateB + rateC + 3*rateE + rateF
        expected_true_rate = rateA + rateB + rateC + rateE + rateF

        assert(model.get_total_event_rate() == expected_rate)
        assert(model.get_true_total_event_rate() == expected_true_rate)

    def test_conditional_transitions(self):
        S, I, A, B, C0, C1, D, E, F = "S I A B C0 C1 D E F".split(" ")

        rateA = 3.0
        rateB = 2.0
        rateE = 1.0

        probAC0 = 0.2
        probAC1 = 0.8

        probBD = 0.2

        N = 6
        edges = [ (0, i, 1.0) for i in range(1,N) ]

        model = StochasticEpiModel([S,I,A,B,C0,C1,D, E, F], N, edges)

        model.set_node_transition_processes([
                (I, rateA, A),
                (I, rateB, B),
                (I, rateE, E),
            ])

        model.set_conditional_link_transmission_processes({
            (I, "->", A) : [
                    ( A, S, probAC0, A, C0),
                    ( A, S, probAC1, A, C1),
                ],
            (I, "->", B): [
                    ( B, S, probBD, B, D),
                ],
            (I, "->", E): [
                    ( E, S, "->", E, F),
                ],
            })

        statuses = np.zeros(N,dtype=int)
        statuses[0] = 1

        model.set_node_statuses(statuses)

        counts = np.zeros(model.N_comp,dtype=int)

        N_measurements = 10000

        for meas in range(N_measurements):
            model.set_node_statuses(statuses)

            _ = model.simulate(1e9)
            for c in range(model.N_comp):

                counts[c] += np.count_nonzero(model.node_status == c)

        expected_counts = np.zeros_like(counts)
        expected_counts[model.get_compartment_id(A)] = N_measurements * rateA / (rateB + rateE + rateA)
        expected_counts[model.get_compartment_id(B)] = N_measurements * rateB / (rateB + rateE + rateA)
        expected_counts[model.get_compartment_id(E)] = N_measurements * rateE / (rateB + rateE + rateA)

        expected_counts[model.get_compartment_id(C0)] = N_measurements * ((N-1)*rateA / (rateB + rateE + rateA) * probAC0)
        expected_counts[model.get_compartment_id(C1)] = N_measurements * ((N-1)*rateA / (rateB + rateE + rateA) * probAC1)

        expected_counts[model.get_compartment_id(D)] = N_measurements * ((N-1)*rateB / (rateB + rateE + rateA) * probBD)

        expected_counts[model.get_compartment_id(F)] = (N-1) * expected_counts[model.get_compartment_id(E)]

        expected_counts[model.get_compartment_id(S)] = N_measurements * N - expected_counts.sum()

        ndx = np.where(expected_counts==0)
        counts[ndx] = 1
        expected_counts[ndx] = 1

        _counts = np.delete(counts,1)
        _exp_counts = np.delete(expected_counts,1)
        ndx = np.where(_exp_counts == 0)
        assert(np.all(_counts[ndx]==0))
        ndx = np.where(_exp_counts > 0)
        rel_err = np.abs(1-_counts[ndx]/_exp_counts[ndx])
        assert(rel_err.mean() < 0.06)


    def test_conditional_transmissions(self):
        S, S0,S1, I, A, B, C0, C1, D, E, F = "S S0 S1 I A B C0 C1 D E F".split(" ")

        rateA = 3.0
        rateB = 2.0
        rateE = 1.0

        probAC0 = 0.2
        probAC1 = 0.8

        probBD = 0.2

        N = 6
        edges = [ (0, i, 1.0) for i in range(1,N) ]
        edges.extend([(N+1, 0, 1.0),(N,0,1.0)])

        model = StochasticEpiModel([S,S0,S1,I,A,B,C0,C1,D, E, F], N+2, edges)

        model.set_link_transmission_processes([
                (S0, I, rateA, S0, A),
                (S0, I, rateB, S0, B),
                (S1, I, rateE, S1, E),
            ])

        model.set_conditional_link_transmission_processes({
            (S0, I, "->", S0, A) : [
                    ( A, S, probAC0, A, C0),
                    ( A, S, probAC1, A, C1),
                ],
            (S0, I, "->", S0, B): [
                    ( B, S, probBD, B, D),
                ],
            (S1, I, "->", S1, E): [
                    ( E, S, "->", E, F),
                ],
            })

        statuses = np.zeros(N+2,dtype=int)
        statuses[0] = model.get_compartment_id(I)
        statuses[N+1] = model.get_compartment_id(S0)
        statuses[N] = model.get_compartment_id(S1)

        model.set_node_statuses(statuses)

        counts = np.zeros(model.N_comp,dtype=int)

        N_measurements = 10000

        for meas in range(N_measurements):
            model.set_node_statuses(statuses)

            _ = model.simulate(1e9)

            for c in range(model.N_comp):

                counts[c] += np.count_nonzero(model.node_status == c)


        expected_counts = np.zeros_like(counts)
        expected_counts[model.get_compartment_id(A)] = N_measurements * rateA / (rateB + rateE + rateA)
        expected_counts[model.get_compartment_id(B)] = N_measurements * rateB / (rateB + rateE + rateA)
        expected_counts[model.get_compartment_id(E)] = N_measurements * rateE / (rateB + rateE + rateA)

        expected_counts[model.get_compartment_id(C0)] = N_measurements * ((N-1)*rateA / (rateB + rateE + rateA) * probAC0)
        expected_counts[model.get_compartment_id(C1)] = N_measurements * ((N-1)*rateA / (rateB + rateE + rateA) * probAC1)

        expected_counts[model.get_compartment_id(D)] = N_measurements * ((N-1)*rateB / (rateB + rateE + rateA) * probBD)

        expected_counts[model.get_compartment_id(F)] = (N-1) * expected_counts[model.get_compartment_id(E)]

        expected_counts[model.get_compartment_id(S)] = N_measurements * N - expected_counts.sum()
        expected_counts[model.get_compartment_id(S0)] = N_measurements
        expected_counts[model.get_compartment_id(S1)] = N_measurements

        ndx = np.where(expected_counts==0)
        counts[ndx] = 1
        expected_counts[ndx] = 1

        _counts = np.delete(counts,1)
        _exp_counts = np.delete(expected_counts,1)
        
        ndx = np.where(_exp_counts == 0)
        assert(np.all(_counts[ndx]==0))
        ndx = np.where(_exp_counts > 0)
        rel_err = np.abs(1-_counts[ndx]/_exp_counts[ndx])
        assert(rel_err.mean() < 0.03)

    def test_sampling_callback(self):
        epi = StochasticSIModel(100,infection_rate=5.0)
        epi.set_random_initial_conditions({"S":90,"I":10})
        self.assertRaises(ValueError,epi.simulate,1,sampling_callback=lambda x: x)

        i = 0
        samples = []
        def sampled():
            samples.append(epi.y0[0])

        t, res = epi.simulate(10,sampling_dt=0.1,sampling_callback=sampled)

        assert(all([a==b for a, b in zip(res['S'], samples)]))


    def test_total_event_rate_well_mixed(self):
        N = 4
        I0 = 2
        S0 = N-I0
        k0 = 1
        R0 = 10.0
        rho = 1.0
        eta = R0*rho
        epi = StochasticSIRModel(N,R0=R0,recovery_rate=rho,well_mixed_mean_contact_number=k0)
        epi.set_random_initial_conditions({"S":S0,"I":I0})
        SI_rate = I0*eta*S0/(N-1)
        I_rate = rho*I0
        print(epi.get_true_total_event_rate(), I_rate + SI_rate)
        print(epi.get_total_event_rate(), I_rate + SI_rate)

        assert(np.isclose(epi.get_true_total_event_rate(), I_rate + SI_rate))
        assert(epi.get_true_total_event_rate() < epi.get_total_event_rate())

    
        # This is for testing whether the true event rate should be corrected with N-1 and not with N
        N_sample =  3000

        count_SI = 0
        taus = []
        SI_taus = []
        for sample in range(N_sample):
            epi.set_random_initial_conditions({"S":S0,"I":I0})
            t, res = epi.simulate(1)
            if res['I'][1] > I0:
                count_SI += 1
                SI_taus.append(t[1]-t[0])
            taus.append(t[1]-t[0])

        SI_rate_2 = I0*eta*S0/N
        I_rate = rho*I0
        #print(count_SI/N_sample, SI_rate/(I_rate+SI_rate), )
        #print(count_SI/N_sample, SI_rate_2/(I_rate+SI_rate_2), )

        from matplotlib import pyplot as pl
        pl.figure()
        counts, bins = np.histogram(taus,bins=100,density=True)
        x = 0.5*(bins[1:]+bins[:-1])
        #pl.hist(taus,bins=50,density=True)
        L = I_rate + SI_rate
        L2 = I_rate + SI_rate_2
        #pl.plot(x,L*np.exp(-L*x))
        #pl.plot(x,L2*np.exp(-L2*x))

        #pl.figure()
        counts, bins = np.histogram(SI_taus,bins=100,density=True)
        x = 0.5*(bins[1:]+bins[:-1])
        #pl.hist(SI_taus,bins=50,density=True)
        L = SI_rate
        L2 = SI_rate_2
        #pl.plot(x,L*np.exp(-L*x))
        #pl.plot(x,L2*np.exp(-L2*x))

        #pl.show()


if __name__ == "__main__":

    T = StochasticEpiTest()
    T.test_total_event_rate_well_mixed()
    T.test_sampling_callback()
    T.test_compartments()
    T.test_mean_contact_number()
    T.test_errors()
    T.test_random_initial_conditions()
    T.test_rate_evaluation()
    T.test_conditional_transitions()
    T.test_conditional_transmissions()



"""
Some network models and layout functions.
"""

import numpy as np

def _edge(u,v,w):
    u, v = sorted([u,v])
    return u, v, w

def _dist(i,j,N):
    return min(np.abs(i-j), N-np.abs(i-j))

def get_2D_lattice_links(N_side,periodic=False,diagonal_links=False):
    """
    Return the links of a square 2D lattice

    Parameters
    ==========
    N_side : int
        Number of nodes per side of the square
    periodic : bool, default = False
        Whether or not to apply periodic boundary conditions
    diagonal_links : bool, default = False
        Whether or not to connect to diagonal neighbors, too.
    """

    links = []
    for i in range(N_side):
        for j in range(N_side):
            u = i*N_side + j
            if periodic or j+1 < N_side:
                v = i*N_side + ( (j+1) % N_side )
                links.append(_edge(u,v,1.0))
            if periodic or i+1 < N_side:
                v = ((i+1) % N_side)*N_side + j
                links.append(_edge(u,v,1.0))

            if diagonal_links:
                if periodic or (j+1 < N_side and i+1 <N_side):
                    v = ((i+1) % N_side)*N_side + ( (j+1) % N_side )
                    links.append(_edge(u,v,1.0))
                if periodic or (j-1 >= 0 and i+1 < N_side):
                    v = ((i+1) % N_side)*N_side + ( (j-1) % N_side )
                    links.append(_edge(u,v,1.0))

    if N_side == 2:
        links = list(set(links))

    return links

def get_grid_layout(N_nodes,edge_weight_tuples=[],windowwidth=400,linkwidth=1):
    """
    Returns a stylized network dictionary that puts
    nodes in a grid layout.

    Parameters
    ==========
    N_nodes : int
        The number of nodes in the network
    edge_weight_tuples : list, default = []
        A list of tuples. Each tuple describes an edge
        with the first entry being the source node index,
        the second entry being the target node indes
        and the third entry being the weight, e.g.

        .. code:: python

            [ (0, 1, 1.0) ]

    windowwidth : float, default = 400
        The width of the network visualization
    linkwidth : float, default = 1.0
        All links get the same width.

    Returns
    =======
    network : dict
        A stylized network dictionary in netwulf-format.
    """

    w = h = windowwidth
    N = N_nodes
    N_side = int(np.ceil(np.sqrt(N)))
    dx = w / N_side
    radius = dx/2

    network = {}
    stylized_network = {
        'xlim': [0, w],
        'ylim': [0, h],
        'linkAlpha': 0.5,
        'nodeStrokeWidth': 0.0001,
    }

    nodes = [ {
                'id': i*N_side + j,
                'x_canvas': (i+0.5)*dx,
                'y_canvas': (j+0.5)*dx,
                'radius': radius
              } for i in range(N_side) for j in range(N_side) ]
    nodes = nodes[:N_nodes]
    links = [ {'source': u, 'target': v, 'width': linkwidth} for u, v, w in edge_weight_tuples ]

    stylized_network['nodes'] = nodes
    stylized_network['links'] = links

    return stylized_network

def get_random_layout(N_nodes,
                      edge_weight_tuples=[],
                      windowwidth=400,
                      linkwidth=1,
                      node_scale_by_degree=0.5,
                      circular=True,
                      ):
    """
    Returns a stylized network dictionary that puts
    nodes in a random layout.

    Parameters
    ==========
    N_nodes : int
        The number of nodes in the network
    edge_weight_tuples : list, default = []
        A list of tuples. Each tuple describes an edge
        with the first entry being the source node index,
        the second entry being the target node index
        and the third entry being the weight, e.g.

        .. code:: python

            [ (0, 1, 1.0) ]

    windowwidth : float, default = 400
        The width of the network visualization
    linkwidth : float, default = 1.0
        All links get the same width.
    node_scale_by_degree : float, default = 0.5
        Scale the node radius by ``degree**node_scale_by_degree``.
        Per default, the node disk area will be
        proportional to the degree. If you want
        all nodes to be equally sized, set
        ``node_scale_by_degree = 0``.
    circular : bool, default = True
        Use a circular or square layout

    Returns
    =======
    network : dict
        A stylized network dictionary in netwulf-format.
    """

    w = h = windowwidth
    N = N_nodes
    N_side = int(np.ceil(np.sqrt(N)))
    dx = w / N_side
    radius = dx/3

    network = {}
    stylized_network = {
        'xlim': [0, w],
        'ylim': [0, h],
        'linkAlpha': 0.5,
        'nodeStrokeWidth': 0.0001,
    }

    degree = np.zeros(N,)
    for u, v, _w in edge_weight_tuples:
        degree[u] += 1
        degree[v] += 1

    median_degree = np.median(degree)
    if median_degree == 0:
        median_degree = 1
    radius_scale = (degree/median_degree)**node_scale_by_degree
    radius_scale[radius_scale==0] = 1.0
    radius = radius_scale * radius

    if not circular:
        pos = (np.random.rand(N, 2) * w).tolist()
    else:
        R = w/2
        r = R*np.sqrt(np.random.random(N,))
        t = 2*np.pi*np.random.random(N,)
        pos = np.zeros((N, 2))
        pos[:,0] = r*np.cos(t) + w/2
        pos[:,1] = r*np.sin(t) + w/2


    nodes = [ {
                'id': i,
                'x_canvas': pos[0],
                'y_canvas': pos[1],
                'radius': radius[i],
              } for i, pos in enumerate(pos)]
    nodes = nodes[:N_nodes]
    links = [ {'source': u, 'target': v, 'width': linkwidth} for u, v, w in edge_weight_tuples ]

    stylized_network['nodes'] = nodes
    stylized_network['links'] = links


    return stylized_network

def get_small_world_layout(
                      N_nodes,
                      edge_weight_tuples=[],
                      windowwidth=400,
                      linkwidth=1,
                      node_scale_by_degree=2,
                      nbounce=20,
                      Rbounce=0.1,
                      R=0.8,
                      ):
    """
    Returns a stylized network dictionary that puts
    nodes in a small-world inspired circular layout.
    The ring of nodes will be drawn bouncy
    to better display connections between nearby regions.
    Nodes that connect far-away regions of a network
    will be displayed more centrally.
    Distance is defined as lattice distance by integer
    node id. Hence, nodes must be integers in [0,N).

    Parameters
    ==========
    N_nodes : int
        The number of nodes in the network
    edge_weight_tuples : list, default = []
        A list of tuples. Each tuple describes an edge
        with the first entry being the source node index,
        the second entry being the target node index
        and the third entry being the weight, e.g.

        .. code:: python

            [ (0, 1, 1.0) ]

    windowwidth : float, default = 400
        The width of the network visualization
    linkwidth : float, default = 1.0
        All links get the same width.
    node_scale_by_degree : float, default = 2
        Scale the node radius by ``degree**node_scale_by_degree``.
        If you want
        all nodes to be equally sized, set
        ``node_scale_by_degree = 0``.
    nbounce : int, default = 20
        How wobbly the outer shell should be.
    Rbounce : int, default = 0.1
        How thick the outer shell should be
        (in units of half window width)
    R : float, default = 0.8
        How large the radius of the whole layout should be
        (in units of half window width)

    Returns
    =======
    network : dict
        A stylized network dictionary in netwulf-format.
    """

    w = h = windowwidth
    N = N_nodes
    N_side = int(np.ceil(np.sqrt(N)))
    dx = w / N_side
    radius = dx/3

    Rbounce /= R
    R = R * w/2
    Rbounce *= R

    network = {}
    stylized_network = {
        'xlim': [0, w],
        'ylim': [0, h],
        'linkAlpha': 0.5,
        'nodeStrokeWidth': 0.0001,
    }

    degree = np.zeros(N,)
    node_distances = [[] for n in range(N)]
    for u, v, _w in edge_weight_tuples:
        node_distances[u].append(_dist(u,v,N))
        node_distances[v].append(_dist(u,v,N))
        degree[u] += 1
        degree[v] += 1

    median_degree = np.median(degree)
    if median_degree == 0:
        median_degree = 1
    radius_scale = (degree/median_degree)**node_scale_by_degree
    radius_scale[radius_scale==0] = 1.0
    radius = radius_scale * radius

    node_distance_maxs = [ np.max(dists+[0]) for dists in node_distances ]
    distance_measure = node_distance_maxs

    dphi = 2*np.pi/N
    maxmean = 3*N/4

    pos = np.zeros((N, 2))

    for n in range(N):
        phi = n*dphi
        rbase = R+Rbounce*np.sin(nbounce*phi)
        r = rbase*(1-distance_measure[n]/maxmean)
        x = r*np.cos(phi) + w/2
        y = r*np.sin(phi) + w/2
        pos[n,:] = x, y

    nodes = [ {
                'id': i,
                'x_canvas': pos[0],
                'y_canvas': pos[1],
                'radius': radius[i],
              } for i, pos in enumerate(pos)]
    nodes = nodes[:N_nodes]
    links = [ {'source': u, 'target': v, 'width': linkwidth} for u, v, w in edge_weight_tuples ]

    stylized_network['nodes'] = nodes
    stylized_network['links'] = links


    return stylized_network




"""
Interactive Jupyter widgets for SymbolicEpiModels.
"""

import copy
from collections import OrderedDict
from math import log10

import numpy as np
import sympy

import ipywidgets as widgets

import matplotlib.pyplot as pl

from epipack.colors import palettes, hex_colors

def get_box_layout():
    """Return default box layout"""
    return widgets.Layout(
        margin='0px 10px 10px 0px',
        padding='5px 5px 5px 5px'
    )

class Range(dict):
    """
    Defines a value range for an interactive linear
    value slider.

    Parameters
    ==========
    min : float
        Minimal value of parameter range
    max : float
        Maximal value of parameter range
    step_count : int, default = 100
        Divide the parameter space into that
        many intervals
    value : float, default = None
        Initial value. If ``None``, defaults to the
        mean of ``min`` and ``max``.
    """

    def __init__(self,
                 min,
                 max,
                 step_count=100,
                 value=None):

        super().__init__()

        assert(max > min)
        assert(step_count>0)

        self['min'] = min
        self['max'] = max
        if value is None:
            self['value'] = 0.5*(max+min)
        else:
            assert(min <= value and max >= value)
            self['value'] = value
        self['step'] = (max-min)/step_count

    def __float__(self):
        return float(self['value'])

    def __add__(self, other):
        return other + float(self)

    def __radd__(self, other):
        return other + float(self)

    def __mul__(self, other):
        return other * float(self)

    def __rmul__(self, other):
        return other * float(self)

    def __truediv__(self, other):
        return float(self) / other

    def __rtruediv__(self, other):
        return other / float(self)

    def __pow__(self, other):
        return float(self)**other

    def __rpow__(self, other):
        return other**float(self)

    def __sub__(self, other):
        return float(self) - other

    def __rsub__(self, other):
        return other - float(self)

class LogRange(dict):
    """
    Defines a value range for an interactive logarithmic
    value slider.

    Parameters
    ==========
    min : float
        Minimal value of parameter range
    max : float
        Maximal value of parameter range
    step_count : int, default = 100
        Divide the exponent space into that
        many intervals
    base : float, default = 10
        Base of the logarithm
    value : float, default = None
        Initial value. If ``None``, defaults to the
        geometric mean of ``min`` and ``max``.
    """

    def __init__(self,
                 min,
                 max,
                 step_count=100,
                 value=None,
                 base=10,
                 ):

        super().__init__()

        assert(max > min)
        assert(step_count>0)
        assert(base>0)

        def logB(x):
            return np.log(x) / np.log(base)

        self['min'] = logB(min)
        self['max'] = logB(max)

        if value is None:
            self['value'] = np.sqrt(max*min)
        else:
            assert(min <= value and max >= value)
            self['value'] = value

        self['step'] = (logB(max)-logB(min))/step_count
        self['base'] = base

    def __float__(self):
        return float(self['value'])

    def __add__(self, other):
        return other + float(self)

    def __radd__(self, other):
        return other + float(self)

    def __mul__(self, other):
        return other * float(self)

    def __rmul__(self, other):
        return other * float(self)

    def __truediv__(self, other):
        return float(self) / other

    def __rtruediv__(self, other):
        return other / float(self)

    def __pow__(self, other):
        return float(self)**other

    def __rpow__(self, other):
        return other**float(self)

    def __sub__(self, other):
        return float(self) - other

    def __rsub__(self, other):
        return other - float(self)

class InteractiveIntegrator(widgets.HBox):
    """
    An interactive widget that lets you control parameters
    of a SymbolicEpiModel and shows you the output.

    Based on this tutorial: https://kapernikov.com/ipywidgets-with-matplotlib/

    Parameters
    ==========
    model : epipack.symbolic_epi_models.SymbolicEpiModel
        An instance of ``SymbolicEpiModel`` that has been initiated
        with initial conditions
    parameter_values : dict
        A dictionary that maps parameter symbols to single, fixed values
        or ranges (instances of :class:`epipack.interactive.Range` or
        :class:`epipack.interactive.LogRange`).
    t : numpy.ndarray
        The time points over which the model will be integrated
    return_compartments : list, default = None
        A list of compartments that should be displayed.
        If ``None``, all compartments will be displayed.
    return_derivatives : list, default = None
        A list of derivatives that should be displayed
        If ``None``, no derivatives will be displayed.
    figsize : tuple, default = (4,4)
        Width and height of the created figure.
    palette : str, default = 'dark'
        A palette from ``epipack.colors``. Choose from

        .. code:: python

            [ 'dark', 'light', 'dark pastel', 'light pastel',
              'french79', 'french79 pastel', 'brewer light',
              'brewer dark', 'brewer dark pastel', 'brewer light pastel'
            ]
    integrator : str, default = 'dopri5'
        Either ``euler`` or ``dopri5``.
    continuous_update : bool, default = False
        If ``False``, curves will be updated only if the mouse button
        is released. If ``True``, curves will be continuously updated.
    show_grid : bool, default = False
        Whether or not to display a grid

    Attributes
    ==========
    model : epipack.symbolic_epi_models.SymbolicEpiModel
        An instance of ``SymbolicEpiModel`` that has been initiated
        with initial conditions.
    fixed_parameters : dict
        A dictionary that maps parameter symbols to single, fixed values
    t : numpy.ndarray
        The time points over which the model will be integrated
    return_compartments : list
        A list of compartments that will be displayed.
    colors : list
        A list of hexstrings.
    fig : matplotlib.Figure
        The figure that will be displayed.
    ax : matplotlib.Axis
        The axis that will be displayed.
    lines : dict
        Maps compartments to line objects
    children : list
        Contains two displayed boxes (controls and output)
    continuous_update : bool, default = False
        If ``False``, curves will be updated only if the mouse button
        is released. If ``True``, curves will be continuously updated.
    """

    def __init__(self,
                 model,
                 parameter_values,
                 t,
                 return_compartments=None,
                 return_derivatives=None,
                 figsize=(4,4),
                 palette='dark',
                 integrator='dopri5',
                 continuous_update=False,
                 show_grid=False,
                ):

        super().__init__()

        self.model = model
        self.t = np.array(t)
        self.colors = [ hex_colors[colorname] for colorname in palettes[palette] ]
        if return_compartments is None:
            self.return_compartments = self.model.compartments
        else:
            self.return_compartments = return_compartments
        self.return_derivatives = return_derivatives
        self.integrator = integrator
        self.lines = None
        self.continuous_update = continuous_update

        output = widgets.Output()

        with output:
            self.fig, self.ax = pl.subplots(constrained_layout=True, figsize=figsize)
            self.ax.set_xlabel('time')
            self.ax.set_ylabel('frequency')
            self.ax.grid(show_grid)


        self.fig.canvas.toolbar_position = 'bottom'

        # define widgets
        self.fixed_parameters = {}
        self.sliders = {}
        for parameter, value in parameter_values.items():
            self.fixed_parameters[parameter] = float(value)
            if type(value) not in [Range, LogRange]:
                continue
            else:
                these_vals = copy.deepcopy(value)
                these_vals['description'] = r'\(' + sympy.latex(parameter) + r'\)'
                these_vals['continuous_update'] = self.continuous_update
                if type(value) == LogRange:
                    slider = widgets.FloatLogSlider(**these_vals)
                else:
                    slider = widgets.FloatSlider(**these_vals)
            self.sliders[parameter] = slider

        checkb_xscale = widgets.Checkbox(
            value=False,
            description='logscale time',
        )
        checkb_yscale = widgets.Checkbox(
            value=False,
            description='logscale frequency',
        )

        controls = widgets.VBox(
            list(self.sliders.values()) + [
                checkb_xscale,
                checkb_yscale,
        ])
        controls.layout = get_box_layout()

        out_box = widgets.Box([output])
        output.layout = get_box_layout()

        for parameter, slider in self.sliders.items():
            slider.observe(self.update_parameters, 'value')
        checkb_xscale.observe(self.update_xscale, 'value')
        checkb_yscale.observe(self.update_yscale, 'value')

        self.children = [controls, output]

        self.update_parameters()

    def update_parameters(self, *args, **kwargs):
        """Update the current values of parameters as given by slider positions."""
        parameters = copy.deepcopy(self.fixed_parameters)
        for parameter, slider in self.sliders.items():
            parameters[parameter] = slider.value

        self.update_plot(parameters)

    def update_plot(self, parameters):
        """Recompute and -draw the epidemic curves with updated parameter values"""

        self.model.set_parameter_values(parameters)

        if self.return_derivatives is None:
            res = self.model.integrate(
                    self.t,
                    return_compartments=self.return_compartments,
                    integrator=self.integrator)

        else:
            res = self.model.integrate_and_return_by_index(
                    self.t,
                    integrator=self.integrator)
            ndx = [ self.model.get_compartment_id(C) for C in self.return_derivatives ]
            dydt = self.model.get_numerical_dydt()
            derivatives = np.array([ dydt(t,res[:,it]) for it, t in enumerate(self.t) ]).T
            res = {C: res[self.model.get_compartment_id(C),:] for C in self.return_compartments}
            der = {C: derivatives[self.model.get_compartment_id(C),:] for C in self.return_derivatives}

        is_initial_run = self.lines is None
        if is_initial_run:
            self.lines = {}

        # plot compartments 
        for iC, C in enumerate(self.return_compartments):
            ydata = res[C]
            if is_initial_run:
                self.lines[C], = self.ax.plot(self.t,ydata,label=str(C),color=self.colors[iC])
            else:
                self.lines[C].set_ydata(ydata)

        # plot derivatives
        if self.return_derivatives is not None:

            for iC, C in enumerate(self.return_derivatives):
                ydata = der[C]
                _C = 'd' + str(C) + '/dt'
                if is_initial_run:
                    self.lines[_C], = self.ax.plot(self.t,ydata,ls='--',label=_C,color=self.colors[iC])
                else:
                    self.lines[_C].set_ydata(ydata)


        if is_initial_run:
            self.ax.legend()

        self.fig.canvas.draw()

    def update_xscale(self, change):
        """Update the scale of the x-axis. For "log", pass an object ``change`` that has ``change.new=True``"""
        scale = 'linear'
        if change.new:
            scale = 'log'
        self.ax.set_xscale(scale)

    def update_yscale(self, change):
        """Update the scale of the y-axis. For "log", pass an object ``change`` that has ``change.new=True``"""
        scale = 'linear'
        if change.new:
            scale = 'log'
        self.ax.set_yscale(scale)

class GeneralInteractiveWidget(widgets.HBox):
    """
    An interactive widget that lets you control parameters
    that are passed to a custom function which returns a result
    dictionary.

    Based on this tutorial: https://kapernikov.com/ipywidgets-with-matplotlib/

    Parameters
    ==========
    result_function : func
        A function that returns a result dictionary when passed
        parameter values as ``result_function(**parameter_values)``.
    parameter_values : dict
        A dictionary that maps parameter names to single, fixed values
        or ranges (instances of :class:`epipack.interactive.Range` or
        :class:`epipack.interactive.LogRange`).
    t : numpy.ndarray
        The time points corresponding to values in the result dictionary.
    return_keys : list, default = None
        A list of result keys that should be shown.
        If ``None``, all compartments will be displayed.
    figsize : tuple, default = (4,4)
        Width and height of the created figure.
    palette : str, default = 'dark'
        A palette from ``epipack.colors``. Choose from

        .. code:: python

            [ 'dark', 'light', 'dark pastel', 'light pastel',
              'french79', 'french79 pastel', 'brewer light',
              'brewer dark', 'brewer dark pastel', 'brewer light pastel'
            ]
    continuous_update : bool, default = False
        If ``False``, curves will be updated only if the mouse button
        is released. If ``True``, curves will be continuously updated.
    show_grid : bool, default = False
        Whether or not to display a grid
    ylabel : str, default = 'frequency'
        What to name the yaxis
    label_converter : func, default = str
        A function that returns a string when passed a result key
        or parameter name.


    Attributes
    ==========
    result_function : func
        A function that returns a result dictionary when passed
        parameter values as ``result_function(**parameter_values)``.
    fixed_parameters : dict
        A dictionary that maps parameter names to fixed values
    t : numpy.ndarray
        The time points corresponding to values in the result dictionary.
    return_keys : list
        A list of result dictionary keys of which the result
        will be displayed.
    colors : list
        A list of hexstrings.
    fig : matplotlib.Figure
        The figure that will be displayed.
    ax : matplotlib.Axis
        The axis that will be displayed.
    lines : dict
        Maps compartments to line objects
    children : list
        Contains two displayed boxes (controls and output)
    continuous_update : bool, default = False
        If ``False``, curves will be updated only if the mouse button
        is released. If ``True``, curves will be continuously updated.
    lbl : func, default = str
        A function that returns a string when passed a result key
        or parameter name.
    """

    def __init__(self,
                 result_function,
                 parameter_values,
                 t,
                 return_keys=None,
                 figsize=(4,4),
                 palette='dark',
                 continuous_update=False,
                 show_grid=False,
                 ylabel='frequency',
                 label_converter=str,
                ):

        super().__init__()

        self.t = t
        self.get_result = result_function
        self.colors = [ hex_colors[colorname] for colorname in palettes[palette] ]
        self.return_keys = return_keys
        self.lines = None
        self.continuous_update = continuous_update
        self.lbl = label_converter
        output = widgets.Output()

        with output:
            self.fig, self.ax = pl.subplots(constrained_layout=True, figsize=figsize)
            self.ax.set_xlabel('time')
            self.ax.set_ylabel(ylabel)
            self.ax.grid(show_grid)

        self.fig.canvas.toolbar_position = 'bottom'

        # define widgets
        self.fixed_parameters = {}
        self.sliders = {}
        for parameter, value in parameter_values.items():
            self.fixed_parameters[parameter] = float(value)
            if type(value) not in [Range, LogRange]:
                continue
            else:
                these_vals = copy.deepcopy(value)
                these_vals['description'] = self.lbl(parameter) or parameter
                these_vals['continuous_update'] = self.continuous_update
                if type(value) == LogRange:
                    slider = widgets.FloatLogSlider(**these_vals)
                else:
                    slider = widgets.FloatSlider(**these_vals)
            self.sliders[parameter] = slider

        checkb_xscale = widgets.Checkbox(
            value=False,
            description='logscale time',
        )
        checkb_yscale = widgets.Checkbox(
            value=False,
            description='logscale frequency',
        )

        controls = widgets.VBox(
            list(self.sliders.values()) + [
                checkb_xscale,
                checkb_yscale,
        ])
        controls.layout = get_box_layout()

        out_box = widgets.Box([output])
        output.layout = get_box_layout()

        for parameter, slider in self.sliders.items():
            slider.observe(self.update_parameters, 'value')
        checkb_xscale.observe(self.update_xscale, 'value')
        checkb_yscale.observe(self.update_yscale, 'value')

        self.children = [controls, output]

        self.update_parameters()

    def update_parameters(self, *args, **kwargs):
        """Update the current values of parameters as given by slider positions."""
        parameters = copy.deepcopy(self.fixed_parameters)
        for parameter, slider in self.sliders.items():
            parameters[parameter] = slider.value

        self.update_plot(parameters)

    def update_plot(self, parameters):
        """Recompute and -draw the epidemic curves with updated parameter values"""

        res = self.get_result(**parameters)

        is_initial_run = self.lines is None
        if is_initial_run:
            self.lines = {}

        if self.return_keys is None:
            keys = res.keys()
        else:
            keys = self.return_keys

        # plot compartments 
        for iC, C in enumerate(keys):
            ydata = res[C]
            if is_initial_run:
                self.lines[C], = self.ax.plot(self.t,ydata,label=self.lbl(C),color=self.colors[iC])
            else:
                self.lines[C].set_ydata(ydata)

        if is_initial_run:
            self.ax.legend()

        self.fig.canvas.draw()

    def update_xscale(self, change):
        """Update the scale of the x-axis. For "log", pass an object ``change`` that has ``change.new=True``"""
        scale = 'linear'
        if change.new:
            scale = 'log'
        self.ax.set_xscale(scale)

    def update_yscale(self, change):
        """Update the scale of the y-axis. For "log", pass an object ``change`` that has ``change.new=True``"""
        scale = 'linear'
        if change.new:
            scale = 'log'
        self.ax.set_yscale(scale)


if __name__=="__main__": # pragma: no cover

    A = LogRange(0.1,1,value=0.5)

    print(A + 2)
    print(2 + A)
    print(A * 2)
    print(2 * A)
    print(A / 2)
    print(2 / A)
    print(A**2)
    print(2**A)
    print(A - 2)
    print(2 - A)



import unittest

import numpy as np
import sympy
from sympy import symbols, Matrix, Eq, sqrt, FiniteSet, Derivative

from scipy.optimize import root

from epipack import (
            SymbolicMatrixSISModel,
            SymbolicMatrixSIRModel,
            SymbolicMatrixSIRSModel,
            SymbolicMatrixSIModel,
            MatrixSISModel,
            MatrixEpiModel,
            SymbolicMatrixEpiModel,
        )

class SymbolicMatrixEpiTest(unittest.TestCase):

    def test_compartments(self):
        comps = sympy.symbols("S E I R")
        epi = SymbolicMatrixEpiModel(comps)
        assert(all([ i == epi.get_compartment_id(C) for i, C in enumerate(comps) ]))

    def test_linear_rates(self):

        S, E, I, R = sympy.symbols("S E I R")
        epi = SymbolicMatrixEpiModel([S,E,I,R])
        epi.add_transition_processes([
                (E, 1, I),
                (I, 1, R),
                ])

        linear_rates = sympy.zeros(4,4)
        linear_rates[1,1] = -1
        linear_rates[2,1] = +1
        linear_rates[2,2] = -1
        linear_rates[3,2] = +1

        N = epi.N_comp
        assert(all([epi.linear_rates[i,j]==linear_rates[i,j] for i in range(N) for j in range(N)]))

    def test_adding_linear_rates(self):

        S, E, I, R = sympy.symbols("S E I R")
        epi = SymbolicMatrixEpiModel([S,E,I,R])

        epi.set_processes([
                (E, 1, I),
                ])

        epi.add_transition_processes([
                (I, 1, R),
                ])

        linear_rates = sympy.zeros(4,4)
        linear_rates[1,1] = -1
        linear_rates[2,1] = +1
        linear_rates[2,2] = -1
        linear_rates[3,2] = +1

        N = epi.N_comp
        assert(all([epi.linear_rates[i,j]==linear_rates[i,j] for i in range(N) for j in range(N)]))

    def test_quadratic_processes(self):

        S, E, I, R = sympy.symbols("S E I R")
        epi = SymbolicMatrixEpiModel([S,E,I,R])
        Q = [ sympy.zeros(4,4) for C in epi.compartments ]
        Q[0][0,2] = -1
        Q[1][0,2] = +1
        epi.add_transmission_processes([
                (S, I,  1, I, E),
                ])
        N = epi.N_comp
        for iM, M in enumerate(epi.quadratic_rates):
            assert(all([M[i,j]==Q[iM][i,j] for i in range(N) for j in range(N)]))

    def test_adding_quadratic_processes(self):

        S, E, I, A, R, rho = sympy.symbols("S E I A R rho")
        epi = SymbolicMatrixEpiModel([S,E,I,A,R])

        Q = [ sympy.zeros(5,5) for C in epi.compartments ]
        Q[0][0,2] = -rho
        Q[0][0,3] = -rho
        Q[1][0,2] = +rho
        Q[1][0,3] = +rho
        epi.set_processes([
                (S, I, rho, I, E),
                ])
        epi.add_transmission_processes([
                (S, A, rho, A, E),
                ])
        N = epi.N_comp
        for iM, M in enumerate(epi.quadratic_rates):
            assert(all([M[i,j]==Q[iM][i,j] for i in range(N) for j in range(N)]))


    def test_basic_analytics(self):
        S, I, R, eta, rho, omega, t = symbols("S I R eta rho omega, t")

        SIRS = SymbolicMatrixEpiModel([S,I,R])

        SIRS.set_processes([
            #### transmission process ####
            # S + I (eta)-> I + I
            (S, I, eta, I, I),

            #### transition processes ####
            # I (rho)-> R
            # R (omega)-> S
            (I, rho, R),
            (R, omega, S),

        ])

        odes = SIRS.ODEs()

        expected = [Eq(Derivative(S, t), -I*S*eta + R*omega), Eq(Derivative(I, t), I*(S*eta - rho)), Eq(Derivative(R, t), I*rho - R*omega)]

        assert(all([got==exp for got, exp in zip(odes,expected) ]))

        fixed_points = SIRS.find_fixed_points()
        expected = FiniteSet((S, 0, 0), (rho/eta, R*omega/rho, R))

        assert(all([got==exp for got, exp in zip(fixed_points,expected) ]))

        J = SIRS.jacobian()
        expected = Matrix([[-I*eta, -S*eta, omega], [I*eta, S*eta - rho, 0], [0, rho, -omega]])
        N = SIRS.N_comp
        assert(all([J[i,j]==expected[i,j] for i in range(N) for j in range(N)]))

        eig = SIRS.get_eigenvalues_at_disease_free_state()
        expected = {-omega: 1, eta - rho: 1, 0: 1}

        assert(all([v == expected[k] for k, v in eig.items()]))

    def test_functional_rates(self):

        u, v, k, f, t = symbols("u v k f t")

        GS = SymbolicMatrixEpiModel([u,v])

        GS.set_processes([
            # third-order coupling
            (u, v, v, v, v),
            # birth and death
            (None, f, u),
            (u, f, None),
            (v, f+k, None),
        ],ignore_rate_position_checks=True)


        odes = GS.ODEs()
        expected = [Eq(Derivative(u, t), -f*u + f - u*v**2), Eq(Derivative(v, t), v*(-f - k + u*v))]
        assert(all([got==exp for got, exp in zip(odes,expected) ]))

        fixed_points = GS.find_fixed_points()
        expected = FiniteSet((1, 0), (-(-f + (f + k)*(f/(2*(f + k)) - sqrt(-f*(4*f**2 + 8*f*k - f + 4*k**2))/(2*(f + k))))/f, f/(2*(f + k)) - sqrt(-f*(4*f**2 + 8*f*k - f + 4*k**2))/(2*(f + k))), (-(-f + (f + k)*(f/(2*(f + k)) + sqrt(-f*(4*f**2 + 8*f*k - f + 4*k**2))/(2*(f + k))))/f, f/(2*(f + k)) + sqrt(-f*(4*f**2 + 8*f*k - f + 4*k**2))/(2*(f + k))))

        evals_a, evals_b = [], []
        for got, exp in zip(fixed_points,expected):
            for a, b in zip(got, exp): 
                eval_a = a.evalf(subs={u:0.1,v:0.2,f:0.3,k:0.4})
                eval_b = b.evalf(subs={u:0.1,v:0.2,f:0.3,k:0.4})
                evals_a.append(eval_a)
                evals_b.append(eval_b)

        assert(set(evals_a) == set(evals_b))

        J = GS.jacobian()
        expected = Matrix([[-f - v**2, -2*u*v], [v**2, -f - k + 2*u*v]])
        N = GS.N_comp
        assert(all([J[i,j]==expected[i,j] for i in range(N) for j in range(N)]))

        eig = GS.get_eigenvalues_at_disease_free_state({"u":1})
        expected = {-f: 1, -f - k: 1}
        assert(all([v == expected[k] for k, v in eig.items()]))

    def test_numerical(self):

        S, I, eta, rho = sympy.symbols("S I eta rho")

        epi = SymbolicMatrixSISModel(eta, rho)

        epi.set_initial_conditions({S: 1-0.01, I:0.01 })
        epi.set_parameter_values({eta:2})

        self.assertRaises(ValueError, epi.integrate,
                    [0,1]
                    )

        epi = SymbolicMatrixSISModel(eta, rho)

        epi.set_initial_conditions({S: 1-0.01, I:0.01 })
        epi.set_parameter_values({eta:2,rho:1})

        tt = np.linspace(0,10,1000)
        result = epi.integrate(tt)

        epi2 = MatrixSISModel(2, 1)

        epi2.set_initial_conditions({"S": 1-0.01, "I":0.01 })

        tt = np.linspace(0,10,1000)
        result2 = epi2.integrate(tt)

        for c0, res0 in result.items():
            assert(np.allclose(res0, result2[str(c0)]))


    def test_time_dependent_rates(self):

        B, t = sympy.symbols("B t")
        epi = SymbolicMatrixEpiModel([B])
        epi.add_fission_processes([
                (B, t, B, B),
            ])
        epi.set_initial_conditions({B:1})
        result = epi.integrate([0,3],adopt_final_state=True)
        assert(np.isclose(epi.y0[0], np.exp(3**2/2)))
        epi.set_initial_conditions({B:1})
        result = epi.integrate(np.linspace(0,3,10000),integrator='euler',adopt_final_state=True)
        eul = epi.y0[0]
        real = np.exp(3**2/2)
        assert(np.abs(1-eul/real)<1e-2)

    def test_exceptions(self):

        B, mu, t = sympy.symbols("B mu t")
        epi = SymbolicMatrixEpiModel([B])
        epi.add_fission_processes([
                (B, mu, B, B),
            ])
        epi.set_initial_conditions({B:1})

        self.assertRaises(ValueError,epi.integrate,[0,1])

        self.assertRaises(ValueError,SymbolicMatrixEpiModel,[t])

        self.assertRaises(ValueError,epi.get_eigenvalues_at_disease_free_state)

    def test_custom_models(self):
        eta = sympy.symbols("eta")
        epi = SymbolicMatrixSIModel(eta)
        epi.set_parameter_values({eta:1})
        epi.set_initial_conditions({epi.compartments[0]:0.99, epi.compartments[1]:0.01})
        epi.integrate([0,1000],adopt_final_state=True)
        assert(np.isclose(epi.y0[0],0))


        S, I, R, eta, rho = sympy.symbols("S I R eta rho")
        epi = SymbolicMatrixSIRModel(eta,rho)
        epi.set_parameter_values({eta:2,rho:1})
        S0 = 0.99
        epi.set_initial_conditions({S:S0, I:1-S0})
        R0 = 2
        Rinf = lambda x: 1-x-S0*np.exp(-x*R0)
        res = epi.integrate([0,1000])

        theory = root(Rinf,0.5)
        assert(np.isclose(res[R][-1],theory.x[0]))

        epi = SymbolicMatrixSISModel(eta, rho, initial_population_size=100)

        epi.set_initial_conditions({S: 99, I:1 })
        epi.set_parameter_values({eta:2,rho:1})

        tt = np.linspace(0,1000,2)
        result = epi.integrate(tt)
        assert(np.isclose(result[S][-1],50))

        S, I, R, eta, rho, omega = sympy.symbols("S I R eta rho omega")
        epi = SymbolicMatrixSIRSModel(eta, rho, omega)
        _eta = 2
        _rho = 1
        _omega = 1

        epi.set_initial_conditions({S: 0.99, I:0.01 })
        epi.set_parameter_values({eta:_eta, rho:_rho, omega:_omega})

        tt = np.linspace(0,1000,2)
        result = epi.integrate(tt)
        assert(np.isclose(result[R][-1],(1-_rho/_eta)/(1+_omega/_rho)))

    def test_ODEs_jupyter(self):
        
        eta = sympy.symbols("eta")
        epi = SymbolicMatrixSIModel(eta)
        epi.ODEs_jupyter()

    def test_changing_population_size(self):

        A, B, C = sympy.symbols("A B C")
        epi = SymbolicMatrixEpiModel([A,B,C],10,correct_for_dynamical_population_size=True)
        epi.set_processes([
                (A, B, 1, C),
            ])

        dydt = epi.dydt()
        assert(dydt[0] == -1*A*B/(A + B + C))



if __name__ == "__main__":

    T = SymbolicMatrixEpiTest()
    T.test_changing_population_size()
    T.test_compartments()
    T.test_linear_rates()
    T.test_adding_linear_rates()
    T.test_quadratic_processes()
    T.test_adding_quadratic_processes()
    T.test_basic_analytics()
    T.test_functional_rates()
    T.test_numerical()
    T.test_time_dependent_rates()
    T.test_exceptions()
    T.test_custom_models()
    T.test_ODEs_jupyter()



"""
Provides an API to define epidemiological
models in terms of sympy symbolic expressions.
"""

import warnings

import numpy as np
import scipy.sparse as sprs

import sympy

from epipack.numeric_epi_models import EpiModel, custom_choice
from IPython.display import Math, display
from sympy.printing.theanocode import theano_function

class SymbolicMixin():
    """
    Provides methods that are useful to both
    :class:`epipack.symbolic_epi_models.SymbolicEpiModel`
    and
    :class:`epipack.symbolic_matrix_epi_models.SymbolicMatrixEpiModel`
    """

    def ODEs(self):
        """
        Obtain the equations of motion for this model in form of equations.
        """
        t = sympy.symbols("t")
        Eqs = []
        ynew = self.dydt()
        for compartment, expr in zip(self.compartments,ynew):
            dXdt = sympy.Derivative(compartment, t)
            Eqs.append(sympy.Eq(dXdt, sympy.simplify(expr)))
        return Eqs

    def ODEs_jupyter(self):
        """
        Pretty-print the equations of motion for this model in a Jupyter notebook.
        """

        for ode in self.ODEs():
            display(Math(sympy.latex(ode)))

    def find_fixed_points(self):
        """
        Find states in which this model is stationary (fixed points).
        """
        ynew = [expr for expr in self.dydt()]
        return sympy.nonlinsolve(ynew, self.compartments)


    def jacobian(self,simplify=True):
        """
        Obtain the Jacobian for this model.

        Parameters
        ----------
        simplify : bool, default = True
            If ``True``, `epipack` will try to simplify
            the evaluated Jacobian. This might not be
            desirable in some cases due to its
            long evaluation time, which is why
            it can be turned off.
        """

        try:
            self.linear_rates
            has_matrix = True
        except AttributeError as e:
            has_matrix = False

        if has_matrix and not self.has_functional_rates:
            y = sympy.Matrix(self.compartments)
            J = sympy.Matrix(self.linear_rates)

            for i in range(self.N_comp):
                J[i,:] += (self.quadratic_rates[i] * y + self.quadratic_rates[i].T * y).T

        else:
            y = sympy.Matrix(self.compartments)
            J = sympy.zeros(self.N_comp, self.N_comp)
            dydt = self.dydt()
            for i in range(self.N_comp):
                for j in range(self.N_comp):
                    J[i,j] = sympy.diff(dydt[i], self.compartments[j])

        if simplify:
            J = sympy.simplify(J)

        return J

    def get_jacobian_at_fixed_point(self,fixed_point_dict,simplify=True):
        """
        Obtain the Jacobian at a given fixed point.

        Parameters
        ----------
        fixed_point_dict : dict
            A dictionary where a compartment symbol maps to an expression
            (the value of this compartment in the fixed point).
            If compartments are missing, it is implicitly assumed
            that this compartment has a value of zero.
        simplify : bool
            whether or not to let sympy try to simplify the expressions

        Returns
        -------
        J : sympy.Matrix
            The Jacobian matrix at the given fixed point.

        """

        fixed_point = self._convert_fixed_point_dict(fixed_point_dict)

        J = self.jacobian(False)

        for compartment, value in fixed_point:
            J = J.subs(compartment, value)

        if simplify:
            J = sympy.simplify(J)

        return J

    def get_eigenvalues_at_fixed_point(self,fixed_point_dict):
        """
        Obtain the Jacobian's eigenvalues at a given fixed point.

        Parameters
        ----------
        fixed_point_dict : dict
            A dictionary where a compartment symbol maps to an expression
            (the value of this compartment in the fixed point). 
            If compartments are missing, it is implicitly assumed
            that this compartment has a value of zero.

        Returns
        -------
        eigenvalues : dict
            Each entry maps an eigenvalue expression to its multiplicity.
        """
        J = self.get_jacobian_at_fixed_point(fixed_point_dict)
        return J.eigenvals()


    def get_eigenvalues_at_disease_free_state(self,disease_free_state=None):
        """
        Obtain the Jacobian's eigenvalues at the disease free state.

        Parameters
        ----------
        disease_free_state : dict, default = None
            A dictionary where a compartment symbol maps to an expression
            (the value of this compartment in the fixed point).
            If compartments are missing, it is implicitly assumed
            that this compartment has a value of zero.

            If ``None``, the disease_free_state is assumed to be at
            ``disease_free_state = { S: 1 }``.

        Returns
        -------
        eigenvalues : dict
            Each entry maps an eigenvalue expression to its multiplicity.

        """

        if disease_free_state is None:
            S = sympy.symbols("S")
            if S not in self.compartments:
                raise ValueError("The disease free state was not provided to the method. "+\
                                 "I tried to assume the disease free state is at S = 1, "+\
                                 "but no `S`-compartment was found.")
            disease_free_state = {S:1}

        return self.get_eigenvalues_at_fixed_point(disease_free_state)

    def _convert_fixed_point_dict(self,fixed_point_dict):
        """
        Get a fixed point item iterator.

        Parameters
        ----------
        fixed_point_dict : dict
            A dictionary where a compartment symbol maps to an expression
            (the value of this compartment in the fixed point).
            If compartments are missing, it is implicitly assumed
            that this compartment has a value of zero.

        Returns
        -------
        fixed_point : iterator of list of tuple
            A list that is ``N_comp`` entries long.
            Each entry contains a compartment symbol and
            an expression that corresponds to the value
            this compartment assumes in the fixed point.
        """
        fixed_point = {}
        for compartment in self.compartments:
            fixed_point[compartment] = 0
        fixed_point.update(fixed_point_dict)
        return fixed_point.items()

    def set_parameter_values(self,parameter_values):
        """
        Set numerical values for the parameters of this model.
        This might include free symbols that are part of symbolized
        rate functions.

        Parameters
        ==========
        parameter_values : dict
            A dictionary mapping symbols to
            numerical values.
        """
        self.parameter_values = parameter_values

    def get_numerical_dydt(self,lambdify_modules='numpy'):
        r"""
        Returns values of the given compartments at the demanded
        time points (as a numpy.ndarray of shape
        ``(return_compartments), len(time_points)``.

        Parameters
        ==========
        time_points : np.ndarray
            An array of time points at which the compartment values
            should be evaluated and returned.
        return_compartments : list, default = None
            A list of compartments for which the result should be returned.
            If ``return_compartments`` is None, all compartments will
            be returned.
        integrator : str, default = 'dopri5'
            Which method to use for integration. Currently supported are
            ``'euler'`` and ``'dopri5'``. If ``'euler'`` is chosen,
            :math:`\delta t` will be determined by the difference
            of consecutive entries in ``time_points``.
        adopt_final_state : bool, default = False
            Whether or not to adopt the final state of the integration

        Returns
        =======
        dydt : func
            A function ``dydt(t, y, *args, **kwargs)`` that returns
            the numerical momenta of this system at time ``t`` and
            state vector ``y``.
        """

        these_symbols = [sympy.symbols("t")] + self.compartments

        params = list(self.parameter_values.items())
        param_symbols = set([_p[0] for _p in params])

        odes = [ ode.subs(params) for ode in self.dydt() ]

        not_set = []
        for ode in odes:
            not_set.extend(ode.free_symbols)

        not_set = set(not_set) - set(these_symbols)

        if len(not_set) > 0:
            raise ValueError("Parameters", set(not_set), "have not been set. Please set them using",
                             "SymbolicEpiModel.parameter_values()")

        F_sympy = sympy.lambdify(these_symbols, odes, modules=lambdify_modules)

        def dydt(t, y, *args, **kwargs):
            these_args = [t] + y.tolist()
            return np.array(F_sympy(*these_args))

        return dydt

    def get_numerical_event_and_rate_functions(self):
        """
        Converts the symbolic event lists and corresponding
        symbolic rates to functions that return numeric
        event lists and numeric rates based on the current
        time and state vector.

        This function is needed in the
        :class:`epipack.numeric_epi_models.EpiModel`
        base class for stochastic simulations.

        Returns
        -------
        get_event_rates : func
            A function that takes the current time ``t`` and
            state vector ``y``
            and returns numerical event rate lists.
        get_compartment_changes : funx
            A function that takes a numerical list of event ``rates``
            and returns a random event state change vector
            with probability proportional to its entry in ``rates``.
        """
        rates = self.birth_rate_functions + self.linear_rate_functions
        if self.correct_for_dynamical_population_size:
            population_size = sum(self.compartments)
        else:
            population_size = self.initial_population_size
        rates += [ r/population_size for r in self.quadratic_rate_functions ]
        events = self.birth_event_updates + self.linear_event_updates + self.quadratic_event_updates

        these_symbols = [sympy.symbols("t")] + self.compartments

        params = list(self.parameter_values.items())
        param_symbols = set([_p[0] for _p in params])

        rates = [ sympy.sympify(rate).subs(params) for rate in rates ]
        events = [ np.array(event).astype(np.float64).flatten() for event in events ]

        not_set = []
        for rate in rates:
            not_set.extend(rate.free_symbols)

        not_set = set(not_set) - set(these_symbols)

        if len(not_set) > 0:
            raise ValueError("Parameters " + str(set(not_set)) +\
                             "have not been set. Please set them using " +\
                             "SymbolicEpiModel.parameter_values()")

        sympy_rates = sympy.lambdify(these_symbols, rates)

        def get_event_rates(t, y):
            """
            Returns numerical event rate lists based on current
            state vector ``y`` and time ``t``.
            """
            these_args = [t] + y.tolist()
            return np.array(sympy_rates(*these_args))

        def get_compartment_changes(rates):
            """
            Choose an event change vector based on the current ``rates``.
            """
            idy = custom_choice(rates/rates.sum())
            return events[idy]

        return get_event_rates, get_compartment_changes

def get_temporal_interpolation(time_data, value_data, interpolation_degree=1):
    """
    Obtain a symbolic piecewise function that interpolates between values
    given in ``value_data`` for the intervals defined in ``time_data``,
    based on a spline interpolation of degree ``interpolation_degree``.
    If ``interpolation_degree == 0``, the function changes according to step
    functions. In this case ``time_data`` needs to have one value more than
    ``value_data``.

    The values in ``time_data`` and ``value_data`` can be symbols or numeric
    values.

    Parameters
    ----------
    time_data : list
        Sorted list of time values.
    value_data : list
        List of values corresponding to the times given in ``time_data``.
    interpolation_degree : int
        The degree of the polynomial that interpolates between values.
    """
    t = sympy.symbols("t")
    if interpolation_degree == 0:
        if len(time_data) != len(value_data)+1:
            raise ValueError("For ``interpolation_degree == 0``, `time_data`` " +\
                             "needs to have one value more than ``value_data``.")
        return sympy.Piecewise(*[
                    (v, (time_data[i] <= t) & ( t < time_data[i+1])) \
                    for i, v in enumerate(value_data)
                ])
    else:
        return sympy.interpolating_spline(interpolation_degree, t, time_data, value_data)

class SymbolicEpiModel(SymbolicMixin, EpiModel):
    """
    Define a model based on the analytical framework
    offered by `Sympy <https://sympy.org/>`_.

    This class uses the event-based framework
    where state-change vectors are associated with
    event rates.

    Parameters
    ==========
    compartments : list
        A list of :class:`sympy.Symbol` instances that
        symbolize compartments.
    initial_population_size : float, default = 1.0
        The population size at :math:`t = 0`.
    correct_for_dynamical_population_size : bool, default = False
        If ``True``, the quadratic coupling terms will be
        divided by the sum of all compartments, otherwise they
        will be divided by the initial population size.

    Attributes
    ==========
    compartments : list
        A list of :class:`sympy.Symbol` instances that
        symbolize compartments.
    N_comp : :obj:`int`
        Number of compartments (including population number)
    parameter_values : dict
        Maps parameter symbols to numerical values,
    initial_population_size : float
        The population size at :math:`t = 0`.
    correct_for_dynamical_population_size : bool
        If ``True``, the quadratic coupling terms will be
        divided by the sum of all compartments, otherwise they
        will be divided by the initial population size.
    birth_rate_functions : list of symbolic expressions
        A list of functions that return rate values based on time ``t``
        and state vector ``y``. Each entry corresponds to an event update
        in ``self.birth_event_updates``.
    birth_event_updates : list of sympy.Matrix
        A list of vectors. Each entry corresponds to a rate in
        ``birth_rate_functions`` and quantifies the change in
        individual counts in the compartments.
    linear_rate_functions : list of symbolic expressions
        A list of functions that return rate values based on time ``t``
        and state vector ``y``. Each entry corresponds to an event update
        in ``self.linear_event_updates``.
    linear_event_updates : list of sympy.Matrix
        A list of vectors. Each entry corresponds to a rate in
        ``linear_rate_functions`` and quantifies the change in
        individual counts in the compartments.
    quadratic_rate_functions : list of symbolic expressions
        A list of functions that return rate values based on time ``t``
        and state vector ``y``. Each entry corresponds to an event update
        in ``self.quadratic_event_updates``.
    quadratic_event_updates : list of sympy.Matrix
        A list of vectors. Each entry corresponds to a rate in
        ``quadratic_rate_functions`` and quantifies the change in
        individual counts in the compartments.
    y0 : numpy.ndarray
        The initial conditions.
    rates_have_explicit_time_dependence : bool
        Internal switch that's flipped when a non-constant
        rate is passed to the model.


    """

    def __init__(self,compartments,
                      initial_population_size=1,
                      correct_for_dynamical_population_size=False,
                      ):

        EpiModel.__init__(self,
                      compartments,
                      initial_population_size=initial_population_size,
                      correct_for_dynamical_population_size=correct_for_dynamical_population_size,
                )

        self.t = sympy.symbols("t")
        if self.t in self.compartments:
            raise ValueError("Don't use `t` as a compartment symbol, as it is reserved for time.")

        self.parameter_values = {}

    def _check_rate_for_functional_dependency(self,rate):
        """
        Sets the attribute ``rates_have_explicit_time_dependence``
        to ``True`` if ``rate`` has an explicit functional dependency
        on a :class:`sympy.Symbol` called ``"t"``.
        """
        try:
            t = sympy.symbols("t")
            has_time_dependence = t in rate.free_symbols
            self.rates_have_explicit_time_dependence |= has_time_dependence
        except AttributeError as e:
            return

    def set_linear_events(self,event_list,allow_nonzero_column_sums=False,reset_events=True):
        r"""
        Define the linear transition events between compartments.

        Parameters
        ==========
        event_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transition events in the
            following format:

            .. code:: python

                [
                    (
                        ("affected_compartment_0",),
                        rate,
                        [
                            ("affected_compartment_0", dN0),
                            ("affected_compartment_1", dN1),
                            ...
                        ],
                     ),
                    ...
                ]

        allow_nonzero_column_sums : :obj:`bool`, default : False
            Traditionally, epidemiological models preserve the
            total population size. If that's not the case,
            switch off testing for this.
        reset_events : bool, default : True
            Whether to reset all linear events to zero before
            converting those.

        Example
        -------
        For an SEIR model with infectious period ``tau``
        and incubation period ``theta``.

        .. code:: python

            epi.set_linear_events([
                ( ("E",),
                  1/theta, 
                  [ ("E", -1), ("I", +1) ] 
                ),
                ( ("I",),
                  1/tau, 
                  [ ("I", -1), ("R", +1) ] 
                ),
            ])

        Read as "compartment E reacts with rate :math:`1/\theta`
        which leads to the decay of one E particle to one I particle."

        """

        if reset_events:
            birth_rate_functions = []
            birth_event_updates = []
            linear_rate_functions = []
            linear_event_updates = []
        else:
            linear_event_updates = self.linear_event_updates
            birth_event_updates = self.birth_event_updates
            linear_rate_functions = self.linear_rate_functions
            birth_rate_functions = self.birth_rate_functions

        for acting_compartments, rate, affected_compartments in event_list:

            dy = sympy.zeros(self.N_comp,1)
            for trg, change in affected_compartments:
                _t = self.get_compartment_id(trg)
                dy[_t] += change

            if acting_compartments[0] is None:
                self._check_rate_for_functional_dependency(rate)
                birth_event_updates.append( dy )
                birth_rate_functions.append( rate )
            else:
                # check if compartment was defined as a function
                this_compartment = acting_compartments[0]

                self._check_rate_for_functional_dependency(rate)
                this_rate = rate * this_compartment
                linear_event_updates.append( dy )
                linear_rate_functions.append( this_rate )


        if not allow_nonzero_column_sums and len(linear_rate_functions)>0:
            _y = np.ones(self.N_comp)
            test = sympy.zeros(self.N_comp,1)
            for dy, r in zip (linear_event_updates, linear_rate_functions):
                test += dy * r
            for dy, r in zip (birth_event_updates, birth_rate_functions):
                test += dy * r
            test_sum = sum(test)
            if test_sum != 0:
                warnings.warn("events do not sum to zero for each column:" + str(test_sum))

        self.linear_event_updates = linear_event_updates
        self.linear_rate_functions = linear_rate_functions
        self.birth_event_updates = birth_event_updates
        self.birth_rate_functions = birth_rate_functions

        return self

    def set_quadratic_events(self,event_list,allow_nonzero_column_sums=False,reset_events=True):
        r"""
        Define the quadratic transition events between compartments.

        Parameters
        ----------
        event_list : :obj:`list` of :obj:`tuple`
            A list of tuples that contains transmission events in the following format:

            .. code:: python

                [
                    (
                        ("coupling_compartment_0", "coupling_compartment_1"), 
                        rate,
                        [
                            ("affected_compartment_0", dN0),
                            ("affected_compartment_1", dN1),
                            ...
                        ],
                     ),
                    ...
                ]

        allow_nonzero_column_sums : :obj:`bool`, default : False
            Traditionally, epidemiological models preserve the
            total population size. If that's not the case,
            switch off testing for this.
        reset_events : bool, default : True
            Whether to reset all linear events to zero before
            converting those.

        Example
        -------
        For an SEIR model with infection rate ``eta``.

        .. code:: python

            epi.set_quadratic_events([
                ( ("S", "I"),
                  eta, 
                  [ ("S", -1), ("E", +1) ] 
                ),
            ])

        Read  as

        "Coupling of *S* and *I* leads to
        the decay of one *S* particle to one *E* particle with 
        rate :math:`\eta`.".
        """

        if reset_events:
            quadratic_event_updates = []
            quadratic_rate_functions = []
        else:
            quadratic_event_updates = self.quadratic_event_updates
            quadratic_rate_functions = self.quadratic_rate_functions

        for coupling_compartments, rate, affected_compartments in event_list:

            _s0 = coupling_compartments[0]
            _s1 = coupling_compartments[1]

            dy = sympy.zeros(self.N_comp,1)
            for trg, change in affected_compartments:
                _t = self.get_compartment_id(trg)
                dy[_t] += change

            self._check_rate_for_functional_dependency(rate)
            this_rate = rate * _s0 * _s1

            quadratic_event_updates.append( dy )
            quadratic_rate_functions.append( this_rate )

        if not allow_nonzero_column_sums and len(quadratic_rate_functions)>0:
            dy = sympy.zeros(self.N_comp,1)
            test = sympy.zeros(self.N_comp,1)
            for dy, r in zip (quadratic_event_updates, quadratic_rate_functions):
                test += dy * r
            test_sum = sum(test)
            if test_sum != 0:
                warnings.warn("events do not sum to zero for each column:" + str(test_sum))

        self.quadratic_event_updates = quadratic_event_updates
        self.quadratic_rate_functions = quadratic_rate_functions

        return self

    def dydt(self):
        """
        Compute the momenta of the epidemiological model as
        symbolic expressions.

        Parameters
        ----------
        t : :obj:`float`
            Current time
        y : numpy.ndarray
            The entries correspond to the compartment frequencies
            (or counts, depending on population size).
        """
        ynew = sympy.zeros(self.N_comp,1)

        for dy, r in zip(self.birth_event_updates, self.birth_rate_functions):
            ynew += r * dy

        for dy, r in zip(self.linear_event_updates, self.linear_rate_functions):
            ynew += r * dy

        if self.correct_for_dynamical_population_size:
            population_size = sum(self.compartments)
        else:
            population_size = self.initial_population_size

        for dy, r in zip(self.quadratic_event_updates, self.quadratic_rate_functions):
            ynew += r/population_size * dy

        return ynew

class SymbolicODEModel(SymbolicEpiModel):
    """
    Define a model purely based on a list of ODEs.

    Parameters
    ==========
    ODEs : list
        A list of symbolic ODEs in format

        .. code:: python

            sympy.Eq(sympy.Derivative(Y, t), expr)
    """

    def __init__(self, ODEs):

        compartments = self._get_compartments_and_assert_derivatives(ODEs)
        self._dydt = [ eq.rhs for eq in ODEs ]

        SymbolicEpiModel.__init__(self, compartments)

    def _get_compartments_and_assert_derivatives(self, ODEs):
        """
        Iterate through a list of ODES, assert that
        each of them is of format 

        .. code:: python

            sympy.Eq(sympy.Derivative(Y, t), expr)

        and return a list of `Y` symbols that
        are assumed to be compartments (ordered
        by appearance in ``ODEs``).
        """

        t = sympy.symbols("t")
        compartments = []

        for eq in ODEs:

            lhs, rhs = eq.lhs, eq.rhs    
            assert(isinstance(lhs, sympy.Derivative))

            expr = lhs.expr
            variables = lhs.variables
            assert(len(expr.free_symbols) == 1)
            assert(len(variables) == 1)
            assert(variables[0] == t)
            compartments.append(list(expr.free_symbols)[0])

        return compartments

    def dydt(self):
        """
        Return the momenta of the epidemiological model as
        symbolic expressions.
        """

        return self._dydt

    def simulate(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'simulate'")

    def set_linear_events(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'set_linear_events'")

    def set_quadratic_events(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'set_quadratic_events'")

    def set_processes(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'set_processes'")

    def add_transition_processes(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'add_transition_processes'")

    def add_fission_processes(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'add_fission_processes'")

    def add_fusion_processes(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'add_fusion_processes'")

    def add_transmission_processes(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'add_transmission_processes'")

    def add_quadratic_events(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'add_quadratic_events'")

    def add_linear_events(self,*args,**kwargs):
        raise AttributeError("'SymbolicODEModel' object has no attribute 'add_linear_events")


class SymbolicSIModel(SymbolicEpiModel):
    """
    An SI model derived from :class:`epipack.symbolic_epi_models.SymbolicEpiModel`.
    """

    def __init__(self, infection_rate, initial_population_size=1):

        S, I = sympy.symbols("S I")

        SymbolicEpiModel.__init__(self,[S, I], initial_population_size)

        self.set_processes([
                (S, I, infection_rate, I, I),
            ])

class SymbolicSIRModel(SymbolicEpiModel):
    """
    An SIR model derived from :class:`epipack.symbolic_epi_models.SymbolicEpiModel`.
    """

    def __init__(self, infection_rate, recovery_rate, initial_population_size=1):

        S, I, R = sympy.symbols("S I R")

        SymbolicEpiModel.__init__(self,[S, I, R], initial_population_size)

        self.add_transmission_processes([
                (S, I, infection_rate, I, I),
            ])

        self.add_transition_processes([
                (I, recovery_rate, R),
            ])

class SymbolicSISModel(SymbolicEpiModel):
    """
    An SIS model derived from :class:`epipack.symbolic_epi_models.SymbolicEpiModel`.
    """

    def __init__(self, infection_rate, recovery_rate, initial_population_size=1):

        S, I = sympy.symbols("S I")

        SymbolicEpiModel.__init__(self,[S, I], initial_population_size)

        self.add_transmission_processes([
                (S, I, infection_rate, I, I),
            ])

        self.add_transition_processes([
                (I, recovery_rate, S),
            ])

class SymbolicSIRSModel(SymbolicEpiModel):
    """
    An SIRS model derived from :class:`epipack.symbolic_epi_models.SymbolicEpiModel`.
    """

    def __init__(self, infection_rate, recovery_rate, waning_immunity_rate, initial_population_size=1):

        S, I, R = sympy.symbols("S I R")

        SymbolicEpiModel.__init__(self,[S, I, R], initial_population_size)

        self.add_transmission_processes([
                (S, I, infection_rate, I, I),
            ])

        self.add_transition_processes([
                (I, recovery_rate, R),
                (R, waning_immunity_rate, S),
            ])

if __name__=="__main__":    # pragma: no cover
    eta, rho = sympy.symbols("eta rho")
    epi = SymbolicSIRModel(eta, rho)

    #for C, M in zip(epi.compartments, epi.quadratic_rates):
    #    print(C, M)

    print(epi.dydt())
    
    print(epi.jacobian())

    print(epi.get_eigenvalues_at_disease_free_state())

    print(epi.dydt())

    print(epi.ODEs())

    print(epi.find_fixed_points())

    epi = SymbolicSISModel(eta, rho)
    print()
    print(epi.ODEs())
    print(epi.find_fixed_points())

    omega = sympy.symbols("omega")
    epi = SymbolicSIRSModel(eta, rho, omega)
    print()
    print(epi.ODEs())
    print(epi.find_fixed_points())


    import sympy
    from epipack import SymbolicEpiModel

    S, I, eta, rho = sympy.symbols("S I eta rho")

    SIS = SymbolicEpiModel([S,I])
    SIS.add_transmission_processes([
            (I, S, eta, I, I),
        ])
    SIS.add_transition_processes([
            (I, rho, S),
        ])

    print(SIS.find_fixed_points())

    print(SIS.get_eigenvalues_at_fixed_point({S:1}))

    print("==========")
    SIS = SymbolicEpiModel([S,I])
    SIS.set_processes([
            (I, S, eta/(1-I), I, I),
            (I, rho, S),
        ])
    print(SIS.jacobian())
    print(SIS.get_eigenvalues_at_disease_free_state())



    N = sympy.symbols("N")
    epi = SymbolicSIRSModel(eta, rho, omega, initial_population_size=N)
    print()
    print(epi.ODEs())
    print(epi.find_fixed_points())

    print("==========")
    x = sympy.symbols("x")
    SIS = SymbolicEpiModel([x,I])
    SIS.set_processes([
            (I, x, eta/(1-I), I, I),
            (I, rho, x),
        ])
    try:
        print(SIS.get_eigenvalues_at_disease_free_state())
    except ValueError as e:
        print(e)

    print("===========")
    S, I = sympy.symbols("S I")
    epi = SymbolicSISModel(eta, rho)

    epi.set_initial_conditions({S: 1-0.01, I:0.01 })
    epi.set_parameter_values({eta:2,rho:1})

    tt = np.linspace(0,10,1000)
    result = epi.integrate(tt)

    import matplotlib.pyplot as pl

    pl.plot(tt, result[S])
    pl.plot(tt, result[I])

    print("===========")
    t, S, I = sympy.symbols("t S I")
    epi = SymbolicSISModel((1.5+sympy.cos(t))*eta, rho)

    epi.set_initial_conditions({S: 1-0.2, I:0.2 })
    epi.set_parameter_values({eta:2,rho:1})

    tt = np.linspace(0,20,1000)
    result = epi.integrate(tt)

    import matplotlib.pyplot as pl

    pl.figure()
    pl.plot(tt, result[S])
    pl.plot(tt, result[I])

    print("===========")
    t, S, I = sympy.symbols("t S I")
    epi = SymbolicSISModel(eta/(1-S), rho)

    epi.set_initial_conditions({S: 1-0.2, I:0.2 })
    epi.set_parameter_values({eta:2,rho:1})

    tt = np.linspace(0,5,1000)
    result = epi.integrate(tt)

    import matplotlib.pyplot as pl

    pl.figure()
    pl.plot(tt, result[S])
    pl.plot(tt, result[I])
    pl.show()




import unittest

import numpy as np
import sympy

from epipack.interactive import GeneralInteractiveWidget, InteractiveIntegrator, LogRange, Range
from epipack import SymbolicEpiModel

def assert_dicts_equal(a,b):
    for k, v in a.items():
        assert(np.isclose(v, b[k]))

class InteractiveTest(unittest.TestCase):

    def test_ranges(self):
        r = Range(min=-1,max=1,step_count=1000,value=0.2)
        expected = dict(min=-1,max=1,step=2/1000.,value=0.2)

        assert_dicts_equal(r, expected)

        r = Range(min=-1,max=1)
        expected = dict(min=-1,max=1,step=2/100.,value=0)

        assert_dicts_equal(r, expected)

        r = LogRange(min=0.1,max=10,step_count=1000,value=0.2)
        expected = dict(min=-1,max=1,step=2/1000.,value=0.2,base=10)

        assert_dicts_equal(r, expected)

        r = LogRange(min=0.1,max=10)
        expected = dict(min=-1,max=1,step=2/100.,value=1,base=10)

        assert_dicts_equal(r, expected)

        A = LogRange(0.1,1,value=0.5)
        B = Range(0.1,1,value=0.5)
        
        for R in [A, B]:
            assert(np.isclose(R + 2,2.5))
            assert(np.isclose(2 + R,2.5))
            assert(np.isclose(R - 2,-1.5))
            assert(np.isclose(2 - R,1.5))
            assert(np.isclose(R * 2,1.0))
            assert(np.isclose(2 * R,1.0))
            assert(np.isclose(R / 2,0.25))
            assert(np.isclose(2 / R,4.0))
            assert(np.isclose(R**2,0.25))
            assert(np.isclose(2**R,np.sqrt(2)))

    def test_interactive_integrator(self):
        S, I, R, R0, tau, omega = sympy.symbols("S I R R_0 tau omega")

        I0 = 0.01
        model = SymbolicEpiModel([S,I,R])\
             .set_processes([
                    (S, I, R0/tau, I, I),
                    (I, 1/tau, R),
                    (R, omega, S),
                ])\
             .set_initial_conditions({S:1-I0, I:I0})

        parameters = {
            R0: LogRange(min=0.1,max=10,step_count=1000),
            tau: Range(min=0.1,max=10,value=8.0),
            omega: 1/14
        }

        t = np.linspace(0.01,200,1000)
        integrator = InteractiveIntegrator(model, parameters, t, figsize=(3,4), return_compartments=[S,I])

        keys = sorted([ str(k) for k in integrator.sliders.keys() ])
        assert(all([a==b for a, b in zip(sorted(['R_0', 'tau']), keys) ]))

        integrator.update_parameters()

        keys = sorted([ str(k) for k in integrator.lines.keys() ])
        assert(all([ a==b for a, b in zip(sorted(['I', 'S']), keys) ]))

        class change():
            new = True

        integrator.update_xscale(change())
        assert(integrator.ax.get_xscale() == 'log')
        assert(integrator.ax.get_yscale() == 'linear')
        integrator.update_yscale(change())
        assert(integrator.ax.get_xscale() == 'log')
        assert(integrator.ax.get_yscale() == 'log')

        integrator = InteractiveIntegrator(model, parameters, t, figsize=(3,4))
        keys = sorted([ str(k) for k in integrator.lines.keys() ])
        assert(all([ a==b for a, b in zip(sorted([str(C) for C in model.compartments]), keys) ]))


    def test_general_interactive_widget(self):


        t = np.linspace(0,100,1000)
        def get_trig(omega_0,T):
            return {
                'A': np.sin(2*np.pi*t/T+omega_0),
                'B': np.cos(2*np.pi*t/T+omega_0),
            }

        parameters = {
            'omega_0': Range(0,7,100),
            'T': LogRange(10,1e3,100),
        }
        integrator = GeneralInteractiveWidget(get_trig, parameters, t)

        keys = sorted([ str(k) for k in integrator.sliders.keys() ])
        assert(all([a==b for a, b in zip(sorted(['omega_0', 'T']), keys) ]))

        integrator.update_parameters()

        keys = sorted([ str(k) for k in integrator.lines.keys() ])
        assert(all([ a==b for a, b in zip(sorted(['A', 'B']), keys) ]))

        class change():
            new = True

        integrator.update_xscale(change())
        assert(integrator.ax.get_xscale() == 'log')
        assert(integrator.ax.get_yscale() == 'linear')
        integrator.update_yscale(change())
        assert(integrator.ax.get_xscale() == 'log')
        assert(integrator.ax.get_yscale() == 'log')

        integrator = GeneralInteractiveWidget(get_trig, parameters, t, figsize=(3,4))
        keys = sorted([ str(k) for k in integrator.lines.keys() ])
        assert(all([ a==b for a, b in zip(sorted([str(C) for C in ['A','B']]), keys) ]))




if __name__ == "__main__":

    T = InteractiveTest()
    T.test_ranges()
    T.test_interactive_integrator()
    T.test_general_interactive_widget()



# -*- coding: utf-8 -*-
"""
Contains a bunch of information about this package.
"""

__version__ = "0.1.5"

__author__ = "Benjamin F. Maier"
__copyright__ = "Copyright 2020-2021, " + __author__
__credits__ = [__author__]
__license__ = "MIT"
__maintainer__ = __author__
__email__ = "contact@benmaier.org"
__status__ = "Development"



"""
Heuristics to fit duration distributions to sums of
exponentially distributed waiting times.
This module has to be considered work in progress.
It works, but can be much improved, e.g. by
putting conditions on the ordering of waiting
times such that the symmetry in parameter space
can be utilized for fitting.
"""

import numpy as np
from scipy.sparse import csr_matrix
from scipy.integrate import solve_ivp
from scipy.optimize import minimize
from scipy.stats import entropy

_DEFAULT_METHOD = 'RK23'

class ExpChain():
    """
    A class that represents a chain of states
    where the waiting time between consecutive
    states is distributed exponentially with
    predetermined mean. The class can be
    used to fit total waiting time distributions,
    i. e. distributions of sums of exponential
    random variables.

    Parameters
    ==========
    durations : numpy.ndarray of float
        mean waiting times between states in the chain

    Example
    =======

    >>> C = ExpChain([0.2,0.4])
    >>> C.get_mean()
    0.6
    """

    def __init__(self,durations):

        self.n = n = len(durations)
        self.tau = durations
        self.n_st = int(n) + 1
        self.dt = min(durations) / 50.

        data = []
        rows = []
        cols = []

        for i in range(self.n):
            src = i
            trg = i+1
            rows.append(trg)
            cols.append(src)
            data.append(1/self.tau[i])
            rows.append(src)
            cols.append(src)
            data.append(-1/self.tau[i])

        self.W = csr_matrix((data, (rows, cols)), shape=2*[self.n_st])

    def dydt(self,t,y):
        """
        The ODE that performs the transitions between states.
        """
        return self.W.dot(y)

    def get_cdf(self,t=None,percentile_cutoff=0.9999,method=_DEFAULT_METHOD):
        """
        Obtain the cumulative distribution function of the total waiting
        time this chain represents.

        Parameters
        ==========
        t : numpy.ndarray of float, default = None
            Ordered array of time points for which the CDF
            should be returned. If ``None``,
            ``scipy.optimize.solve_ivp`` will choose
            the points itself.
        percentile_cutoff : float, default = 1 - 1e-4
            maximum value of the CDF
        method : str, default = 'RK23'
            This is going to be passed to
            ``scipy.optimize.solve_ivp``.

        Returns
        =======
        t : numpy.ndarray of float
            Ordered array of time points for which the CDF
            was computed.
        cdf : numpy.ndarray of float
            the corresponding values of the cumulative
            distribution function
        """

        y0 = np.zeros(self.n_st)
        y0[0] = 1.0

        stop_condition = lambda t, y: percentile_cutoff - y[-1]
        stop_condition.terminal = True

        result = solve_ivp(
                            fun=self.dydt,
                            y0=y0,
                            t_eval=t,
                            t_span=(0,np.inf),
                            events=stop_condition,
                            method=method,
                          )


        return result.t, result.y[-1,:]

    def get_pdf(self,t=None,percentile_cutoff=0.9999,method=_DEFAULT_METHOD):
        """
        Obtain the probability distribution function of the total waiting
        time this chain represents. Uses :func:`ExpChain.get_cdf`.

        Parameters
        ==========
        t : numpy.ndarray of float, default = None
            Ordered array of time points for which the CDF
            should be returned. If ``None``,
            ``scipy.optimize.solve_ivp`` will choose
            the points itself.
        percentile_cutoff : float, default = 1 - 1e-4
            maximum value of the CDF
        method : str, default = 'RK23'
            This is going to be passed to
            ``scipy.optimize.solve_ivp``.

        Returns
        =======
        tmean : numpy.ndarray of float
            Ordered array of bin midpoints for which the pdf
            was computed.
        pdf : numpy.ndarray of float
            the corresponding values of the pdf.
        df : numpy.ndarray of float
            the corresponding bin sizes
        """
        t, P = self.get_cdf(t=t,percentile_cutoff=percentile_cutoff,method=method)
        dt = np.diff(t)
        dP = np.diff(P)
        tmean = 0.5*(t[1:]+t[:-1])
        return tmean, dP/dt, dt

    def get_cdf_at_percentiles(self,percentiles,method=_DEFAULT_METHOD):
        """
        Obtain the cumulative distribution function of the total waiting
        time this chain represents.

        Parameters
        ==========
        t : numpy.ndarray of float, default = None
            Ordered array of time points for which the CDF
            should be returned. If ``None``,
            ``scipy.optimize.solve_ivp`` will choose
            the points itself.
        percentile_cutoff : float, default = 1 - 1e-4
            maximum value of the CDF
        method : str, default = 'RK23'
            This is going to be passed to
            ``scipy.optimize.solve_ivp``.

        Returns
        =======
        t : numpy.ndarray of float
            Ordered array of time points for which the CDF
            was computed.
        cdf : numpy.ndarray of float
            the corresponding values of the cumulative
            distribution function
        """

        y0 = np.zeros(self.n_st)
        y0[0] = 1.0
        time_values = np.zeros_like(percentiles)
        t = 0

        for i, P in enumerate(percentiles):

            stop_condition = lambda t, y: P - y[-1]
            stop_condition.terminal = True

            result = solve_ivp(
                                fun=self.dydt,
                                y0=y0,
                                t_span=(t,np.inf),
                                events=stop_condition,
                                method=method,
                              )

            y0 = result.y[:,-1]
            t = result.t[-1]
            time_values[i] = t

        return time_values, percentiles

    def get_median_and_iqr(self,method=_DEFAULT_METHOD):
        """
        Returns the median and inter-quartile range
        of the waiting time distribution this chain
        represents.

        Parameters
        ==========
        method : str, default = 'RK23'
            This is going to be passed to
            ``scipy.optimize.solve_ivp``.

        Returns
        =======
        median : float
            the median of the distribution
        iqr : numpy.ndarray of float
            array of length 2 containing
            the inter-quartile range.
        """

        percentiles = [0.25,0.5,0.75]

        time_values, _ = self.get_cdf_at_percentiles(percentiles,method=method)

        return time_values[1], time_values[[0,2]]

    def get_mean(self):
        """
        Returns the mean waiting time of this chain.
        """
        return sum(self.tau)

def fit_chain_by_cdf(n,time_values,cdf,lower=1e-10,upper=1e10,percentile_cutoff=1-1e-15,x0=None):
    """
    Fit a chain of exponentially distributed random variables
    to a distribution where the cdf is known for several time points.

    While there exist statistcally sound measures to quantify
    the distance between
    two distributions, I found that the total mean squared distance
    actually finds decent fits consistently, so this function
    is going to be using that until someone convinces me that
    another distance measure yields better results.

    This whole thing should be considered heuristic
    patch work in any case.

    Parameters
    ==========
    n : int
        number of transitions in the chain
    time_values : numpy.ndarray of float
        Ordered array of time points for which the cdf
        is known.
    cdf : numpy.ndarray of float
        Ordered array of corresponding CDF values.
    lower: float, default = 1e-10
        lower bound of waiting times for each transition
    upper: float, default = 1e10
        upper bound of waiting times for each transition
    percentile_cutoff : float default = 1 - 1e-15
        max value of the CDF that should be integrated to
    x0 : numpy.ndarray of float, default = None
        array of length n that contains initial guesses
        of the chain's waiting times. If ``None``,
        ``x0`` is going to contain the value ``mean/n``
        n times, where ``mean`` is the mean of the
        distribution determined by ``cdf``.

    Returns
    =======
    chain : ExpChain
        The chain that was fit to the given CDF.

    Example
    =======

    >>> median, iqr = 13.184775302968362, ( 7.81098765, 20.86713744)
    >>> fit_C = fit_chain_by_cdf(3,[iqr[0], median, iqr[1]],[0.25,0.5,0.75])
    >>> fit_C.get_median_and_iqr()
    13.183969129892406, array([ 7.8109697 , 20.86699702])
    >>> fit_C.tau
    [9.22794388 0.75881288 5.72462722]

    """

    if n != int(n):
        raise ValueError("`n` must be integer, but is " + str(n))

    dQ = np.diff(cdf)

    cdf = np.array(cdf)

    if x0 is None:
        mean = (np.diff(np.concatenate((np.array([0]),time_values))) * (1 - cdf)).sum()
        x0 = [mean/n] * n

    def cost(x, n, cdf, time_values, percentile_cutoff):
        C  = ExpChain(x)
        _, P = C.get_cdf(time_values,percentile_cutoff=percentile_cutoff)
        if len(P) < len(cdf):
            P = np.concatenate((P,np.ones(len(cdf)-len(P))))
        #return ( ((cdf - P)/cdf)**2).sum()
        return ( ((cdf - P)/cdf)**2).sum()
        #return max(np.abs(cdf - P))
        #return ( ((percentiles - P))**2).sum()
        #dP = np.diff(P)
        #return entropy(np.diff(P), dQ)


    result = minimize(cost, x0, (n, cdf, time_values, percentile_cutoff), bounds=[(lower,upper)]*n)

    return ExpChain(result.x)


def fit_chain_by_median_and_iqr(n,median,iqr,lower=1e-10,upper=1e10):
    """
    Fit a chain of exponentially distributed random variables
    to a distribution where only median and iqr are known.

    Parameters
    ==========
    n : int
        number of transitions in the chain
    median : float
        the median of the distribution to fit to
    iqr : 2-tuple of float
        the inter-quartile range of the distribution to
        fit to
    lower: float, default = 1e-10
        lower bound of waiting times for each transition
    upper: float, default = 1e10
        upper bound of waiting times for each transition

    Returns
    =======
    chain : ExpChain
        The chain that was fit to the median and iqr.

    Example
    =======

    >>> times = [0.3,6.,9,0.4]
    >>> C = ExpChain(times)
    >>> fit_C = fit_chain_by_median_and_iqr(3,*C.get_median_and_iqr())
    >>> C.get_median_and_iqr()
    13.184775302968362, array([ 7.81098765, 20.86713744])
    >>> fit_C.get_median_and_iqr()
    13.183969129892406, array([ 7.8109697 , 20.86699702])
    >>> C.tau
    [0.3, 6.0, 9, 0.4]
    >>> fit_C.tau
    [9.22794388 0.75881288 5.72462722]

    """
    percentiles = np.array([0.25,0.5,0.75])
    time_values = np.array([iqr[0],median,iqr[1]])
    mean = median
    x0 = [mean/n]*n
    return fit_chain_by_cdf(n,time_values,percentiles,lower,upper,x0=x0)


if __name__ == "__main__":
    import matplotlib.pyplot as pl
    from scipy.stats import erlang

    n = 10
    mean = 10
    C = ExpChain([mean/n]*n)
    E = erlang(a=n,scale=mean/n)

    t, y = C.get_cdf()

    pl.plot(t, y)
    pl.plot(t, E.cdf(t))

    print(C.get_median_and_iqr())
    print(E.median(), E.interval(0.5))
    pl.show()


    # =========
    times = [0.3,6.,9,0.4]
    n = len(times)
    C = ExpChain(times)
    fit_C = fit_chain_by_median_and_iqr(3,*C.get_median_and_iqr())

    print("C med iqr =",C.get_median_and_iqr())
    print("fit med iqr =",fit_C.get_median_and_iqr())
    print("C durs =",C.tau)
    print("fit durs =",fit_C.tau)

    fig, ax = pl.subplots(1,2,figsize=(8,4))
    for ch in [C, fit_C]:
        t, P = ch.get_cdf()
        dP = np.diff(P)
        tmean = 0.5*(t[1:] + t[:-1])
        dt = np.diff(t)
        ax[1].bar(tmean, dP/dt,width=dt,alpha=0.5)
        ax[0].plot(t,P)
        print(sum(dP))

    pl.show()


    # ==================

    data = np.array([
    [0.9782608695652173, 0.6117433930093777],
    [1.9565217391304341, 0.7045325376527423],
    [2.934782608695652, 0.7829425973287867],
    [3.9673913043478257, 0.789456048119731],
    [4.945652173913045, 0.8064281992990434],
    [5.923913043478262, 0.7698055792365256],
    [6.9565217391304355, 0.7266458274130909],
    [7.934782608695654, 0.6691081746708346],
    [8.967391304347826, 0.6115693378800797],
    [9.945652173913045, 0.5474957374254049],
    [10.923913043478263, 0.4781933788007957],
    [11.956521739130435, 0.4245761106374918],
    [12.880434782608697, 0.3539677465188974],
    [13.96739130434783, 0.29773491522212747],
    [14.945652173913045, 0.24411883110732224],
    [15.923913043478263, 0.19703869470493518],
    [17.01086956521739, 0.15518494837548535],
    [17.934782608695652, 0.1342497868712702],
    [18.913043478260867, 0.09893435635123615],
    [19.945652173913043, 0.07799682675002373],
    [20.92391304347826, 0.06098204982476085],
    [21.956521739130434, 0.04658046793596648],
    [22.934782608695652, 0.030872880553187487],
    [24.07608695652174, 0.028233636449748856],
    [25, 0.016448801742919406],
    [25.978260869565215, 0.01119873070000943],
    [26.902173913043477, 0.011178601875532879],
    [27.989130434782606, 0.007233352278109395],
    [28.804347826086957, 0.007215591550630007],
    [29.945652173913047, 0.004576347447191376],
    [31.03260869565218, 0.003245476934735203],
    [31.902173913043477, 0.0032265321587572338],
    [32.93478260869565, 0.003204035237283298],
    [33.96739130434783, -0.000740030311641604],
    [35, -0.0007625272331155397],
    [36.141304347826086, 0.0005197972908970172],
    [37.22826086956522, 0.0004961163209245001],
    [38.15217391304348, 0.00047598749644794935],
    [39.07608695652174, 0.001763048214454832],
    [39.94565217391305, -0.0008702756464904482],
    [40.869565217391305, 0.0004167850715164345],
    [41.79347826086956, 0.0017038457895235393],
    [42.93478260869565, -0.0009353983139148703],
    [43.96739130434783, -0.0009578952353888059],
    [44.94565217391305, 0.0003279814341194953],
    ])
    t = np.round(data[:,0])
    print(t)
    P = data[:,1]
    P[P<0] = 0
    P /= P.sum()
    CDF = np.cumsum(P)
    fig, ax = pl.subplots(1,2,figsize=(8,4))
    ax[0].plot(t, CDF)
    ax[1].bar(t, P,alpha=0.4)
    ch = fit_chain_by_cdf(3,t,CDF,lower=0.1,upper=50,x0=[1.29,3.37,4.01])
    #ch = fit_chain_by_cdf(8,t,CDF,lower=0.1,upper=50)
    t, dPdt, width = ch.get_pdf()
    ax[1].bar(t, dPdt, width=width,alpha=0.4)
    ax[0].plot(*ch.get_cdf(t))

    pl.show()



    # ===========

    med_iqr = [
                ( 5, ( 2, 13)),
                (11, ( 6, 21)),
                (10, ( 4, 19)),
                ( 6, ( 3, 11)),
                ( 9, ( 4, 18)),
            ]

    for med, iqr in med_iqr:
        fit_C = fit_chain_by_median_and_iqr(3,med,iqr,lower=0.1,upper=100)
        print("demanded med and iqr =", med, iqr)
        print("fitted   med and iqr =", fit_C.get_median_and_iqr())
        print("fitted durations =",fit_C.tau)
        print()
        fig, ax = pl.subplots(1,2,figsize=(8,4))
        ax[0].plot(*fit_C.get_cdf())
        t, dPdt, width = fit_C.get_pdf()
        ax[1].bar(t, dPdt, width=width)

    pl.show()








import unittest

import numpy as np
from scipy.optimize import root

from epipack.distributions import (
            ExpChain,
            fit_chain_by_cdf,
            fit_chain_by_median_and_iqr,
        )

class DistributionsTest(unittest.TestCase):

    def test_init(self):

        times = [0.3,6.,9,0.4]
        C = ExpChain(times)
        assert(np.isclose(sum(times), C.get_mean()))

    def test_fit(self):

        times = [0.3,6.,9,0.4]
        C = ExpChain(times)
        fit_C = fit_chain_by_median_and_iqr(3,*C.get_median_and_iqr())

        medA, (iqrA0, iqrA1) = C.get_median_and_iqr()
        medF, (iqrF0, iqrF1) = fit_C.get_median_and_iqr()

        assert(np.isclose(medA, medF,rtol=1e-3))
        assert(np.isclose(iqrA0, iqrF0,rtol=1e-3))
        assert(np.isclose(iqrA1, iqrF1,rtol=1e-3))

if __name__=="__main__":
    T = DistributionsTest()
    T.test_init()
    T.test_fit()



import unittest

import numpy as np

from epipack.integrators import integrate_SDE, IntegrationMixin

def rel_err(a,b):
    return np.sqrt(1-a/b)

class IntegratorTest(unittest.TestCase):

    def test_OU(self):

        ym = 3
        f = 2
        D = [0,1.5]
        std = np.sqrt(D[1]**2/2/f)

        t = np.linspace(0,1,100001)

        def dydt(t,y):
            return -f*(y-ym)

        y0 = [ym,ym]
        y = integrate_SDE(dydt, t, y0, D)

        assert(np.all(np.isclose(y[0],ym)))

        _std = np.std(y[1,:])
        _ym = np.mean(y[1,:])

        assert(rel_err(std,_std))
        assert(rel_err(ym,_ym))

    def test_assertions(self):

        A = IntegrationMixin()
        def _():
            return None
        A.get_numerical_dydt = _

        with self.assertRaises(ValueError):
            # should raise f"Unknown integrator"
            A.integrate_and_return_by_index([0],integrator='sdads')

        with self.assertRaises(ValueError):
            # should raise ValueError("'diffusion_constants' undefined but necessary for SDE integration.")
            A.integrate_and_return_by_index([0],integrator='sde')




if __name__ == "__main__":

    T = IntegratorTest()
    T.test_OU()
    T.test_assertions()



"""
Contains integrator methods that will be used
to integrate ODEs or SDEs.
"""

import warnings

import numpy as np
from scipy.integrate import ode, solve_ivp, quad
from scipy.optimize import newton

def integrate_ivp_until(dydt, t0, y0, stop_condition, solve_ivp_kwargs={}):
    """
    Integrate an ODE system with the ``scipy.integrate.solve_ivp`` method.

    Parameters
    ----------
    dydt : function
        A function returning the momenta of the ODE system.
    t0 : float
        Initial time
    y0 : numpy.ndarray
        Initial conditions
    stop_condition : function
        A function ``stop_condition(t,y)`` that flips from
        positive to negative values when a certain
        condition is reached. For instance, when
        an element reaches a certain threshold value:

        .. code:: python

            stop_condition = lambda t, y: 0.3 - y[2]

    solve_ivp_kwargs : dict
        Keyword arguments that will be passed to
        ``scipy.integrate.solve_ivp``.

    Returns
    -------
    t : float
        Times when the system state was recorded
        The final time corresponds to the time when
        the stop condition was reached.
    y : numpy.ndarray
        First axis accesses system states, second
        axis refers to the system state corresponding
        to the time points in ``t``.
    """

    y0 = np.array(y0,dtype=float)
    stop_condition.terminal = True

    result = solve_ivp(
                    fun=dydt,
                    y0=y0,
                    t_span=(t0,np.inf),
                    events=stop_condition,
                    **solve_ivp_kwargs,
                  )

    y = result.y
    t = result.t

    return t, y

def integrate_dopri5(dydt, t, y0, *args):
    """
    Integrate an ODE system with the Runge-Kutte Dormand Prince
    method (with step-size control).

    Parameters
    ----------
    dydt : function
        A function returning the momenta of the ODE system.
    t : numpy.ndarray of float
        Array of time points at which the functions should
        be evaluated.
    y0 : numpy.ndarray
        Initial conditions
    *args : :obj:`list`
        List of parameters that will be passed to the
        momenta function.
    """

    # get copy
    y0 = np.array(y0,dtype=float)
    t = np.array(t,dtype=float)

    t0 = t[0]
    t = t[1:]

    # initiate integrator
    r = ode(dydt)
    r.set_integrator('dopri5')
    r.set_initial_value(y0,t0)
    r.set_f_params(*args)
    result = np.zeros((len(y0),len(t)+1))
    result[:,0] = y0


    # loop through all demanded time points
    for it, t_ in enumerate(t):

            # get result of ODE integration
            y = r.integrate(t_)

            # write result to result vector
            result[:,it+1] = y

    return result

def integrate_euler(dydt, t, y0, *args):
    """
    Integrate an ODE system with Euler's method.

    Parameters
    ----------
    dydt : function
        A function returning the momenta of the ODE system.
    t : numpy.ndarray of float
        Array of time points at which the functions should
        be evaluated.
    y0 : numpy.ndarray
        Initial conditions
    *args : :obj:`list`
        List of parameters that will be passed to the
        momenta function.
    """

    # get copy
    y0 = np.array(y0,dtype=float)
    t = np.array(t,dtype=float)

    t0 = t[0]


    # initiate integrator
    result = np.zeros((len(y0),len(t)))
    result[:,0] = y0
    old_t = t0


    # loop through all demanded time points
    for it, t_ in enumerate(t[1:]):

            # get result of ODE integration
            dt = t_ - old_t
            y = result[:,it] + dt*dydt(old_t, result[:,it], *args)

            # write result to result vector
            result[:,it+1] = y

            old_t = t_

    return result

def integrate_SDE(dydt, t, y0, diffusion_constants, *args):
    """
    Integrate an SDE system with Euler's method.

    Parameters
    ----------
    dydt : function
        A function returning the momenta of the deterministic ODE system.
    t : numpy.ndarray of float
        Array of time points at which the functions should
        be evaluated. Time steps must be equally spaced.
    y0 : numpy.ndarray
        Initial conditions
    diffusion_constants : numpy.ndarray
        Scalar and constant diffusion coefficients as prefactors
        for each compartment's Wiener process (has to be of same
        length as y0)

        corresponds to :math:`D_i` in

        .. math::

                dY_i = f_i(\mathbf Y,t) dt + D_i dW_i
    *args : :obj:`list`
        List of parameters that will be passed to the
        momenta function.
    """

    # get copy
    y0 = np.array(y0,dtype=float)
    t = np.array(t,dtype=float)
    D = np.array(diffusion_constants,dtype=float)

    t0 = t[0]

    dt = t[1]-t[0]
    sqrt_dt = np.sqrt(dt)

    assert(np.all(np.isclose(dt,t[1:]-t[:-1])))
    assert(np.all(D>=0.))
    ndx = np.where(D>0)[0]


    # initiate integrator
    result = np.zeros((len(y0),len(t)))
    result[:,0] = y0

    dW = np.zeros((len(y0),len(t)-1))
    dW[ndx,:] = np.sqrt(dt) * np.random.randn(len(ndx),len(t)-1)

    # loop through all demanded time points
    for it, t_ in enumerate(t[1:]):

            # get result of ODE integration
            y = result[:,it] + dt * dydt(t[it], result[:,it], *args) + D * dW[:,it]

            # write result to result vector
            result[:,it+1] = y

    return result


class IntegrationMixin():
    """
    A helper MixIn that enables the base
    class to set initial conditions
    and integrate numerical ODEs.

    Expects the base class to have the following methods
    and attributes:

    - get_numerical_dydt()
    - get_compartment_id()
    - compartments
    - N_comp
    """

    def integrate_and_return_by_index(self,
                                      time_points,
                                      return_compartments=None,
                                      integrator='dopri5',
                                      adopt_final_state=False,
                                      diffusion_constants=None,
                                      ):
        r"""
        Returns values of the given compartments at the demanded
        time points (as a numpy.ndarray of shape
        ``(return_compartments), len(time_points)``.

        Parameters
        ==========
        time_points : np.ndarray
            An array of time points at which the compartment values
            should be evaluated and returned.
        return_compartments : list, default = None
            A list of compartments for which the result should be returned.
            If ``return_compartments`` is None, all compartments will
            be returned.
        integrator : str, default = 'dopri5'
            Which method to use for integration. Currently supported are
            ``'euler'`` and ``'dopri5'``. If ``'euler'`` is chosen,
            :math:`\delta t` will be determined by the difference
            of consecutive entries in ``time_points``.
        adopt_final_state : bool, default = False
            Whether or not to adopt the final state of the integration
        diffusion_constants : numpy.ndarray
            Scalar and constant diffusion coefficients as prefactors
            for each compartment's Wiener process (has to be of same
            length as y0)

        Returns
        =======
        result : numpy.ndarray
            First axis accesses system states, second
            axis refers to the system state corresponding
            to the time points in ``time_points``.
        """

        dydt = self.get_numerical_dydt()
        self.t0 = time_points[0]

        if integrator.lower() == 'dopri5':
            result = integrate_dopri5(dydt, time_points, self.y0)
        elif integrator.lower() == 'euler':
            result = integrate_euler(dydt, time_points, self.y0)
        elif integrator.lower() == 'sde':
            if diffusion_constants is None:
                raise ValueError("'diffusion_constants' undefined but necessary for SDE integration.")
            result = integrate_euler(dydt, time_points, self.y0, diffusion_constants)
        else:
            raise ValueError(f"Unknown integrator {integrator}")

        if adopt_final_state:
            self.t0 = time_points[-1]
            self.y0 = result[:,-1]

        if return_compartments is not None:
            ndx = [self.get_compartment_id(C) for C in return_compartments]
            result = result[ndx,:]

        return result

    def integrate(self,
                  time_points,
                  return_compartments=None,
                  *args,
                  **kwargs,
                  ):
        r"""
        Returns values of the given compartments at the demanded
        time points (as a dictionary).

        Parameters
        ==========
        time_points : np.ndarray
            An array of time points at which the compartment values
            should be evaluated and returned.
        return_compartments : list, default = None
            A list of compartments for which the result should be returned.
            If ``return_compartments`` is None, all compartments will
            be returned.
        integrator : str, default = 'dopri5'
            Which method to use for integration. Currently supported are
            ``'euler'`` and ``'dopri5'``. If ``'euler'`` is chosen,
            :math:`\delta t` will be determined by the difference
            of consecutive entries in ``time_points``.
        adopt_final_state : bool, default = False
            Whether or not to adopt the final state of the integration
            as new initial conditions.
        diffusion_constants : numpy.ndarray
            Scalar and constant diffusion coefficients as prefactors
            for each compartment's Wiener process (has to be of same
            length as y0)

        Returns
        =======
        result : dict
            Dictionary mapping compartments to time series.
        """
        if return_compartments is None:
            return_compartments = self.compartments

        result = self.integrate_and_return_by_index(time_points, return_compartments,*args,**kwargs)

        result_dict = {}
        for icomp, compartment in enumerate(return_compartments):
            result_dict[compartment] = result[icomp,:]

        return result_dict

    def set_initial_conditions(self, initial_conditions, initial_time=0.0, allow_nonzero_column_sums=False):
        """
        Set the initial conditions for integration

        Parameters
        ----------
        initial_conditions : dict
            A dictionary that maps compartments to a fraction
            of the population. Compartments that are not
            set in this dictionary are assumed to have an initial condition
            of zero.
        allow_nonzero_column_sums : bool, default = False
            If True, an error is raised when the initial conditions do not
            sum to the population size.
        """

        if type(initial_conditions) == dict:
            initial_conditions = list(initial_conditions.items())

        self.y0 = np.zeros(self.N_comp)
        self.t0 = initial_time
        total = 0
        for compartment, amount in initial_conditions:
            total += amount
            if self.y0[self.get_compartment_id(compartment)] != 0:
                warnings.warn('Double entry in initial conditions for compartment '+str(compartment))
            else:
                self.y0[self.get_compartment_id(compartment)] = amount

        if np.abs(total-self.initial_population_size)/self.initial_population_size > 1e-14 and not allow_nonzero_column_sums:
            warnings.warn('Sum of initial conditions does not equal unity.')

        return self

    def integrate_and_return_by_index_until(
                                      self,
                                      t0,
                                      stop_condition,
                                      return_compartments=None,
                                      adopt_final_state=False,
                                      solve_ivp_kwargs={},
                                      ):
        r"""
        Integrate the system until a stop condition is reached,
        using ``scipy.integrate.solve_ivp``.
        Returns values of the given compartments at the demanded
        time points (as a numpy.ndarray of shape
        ``(return_compartments), len(time_points)``.

        Parameters
        ==========
        t0 : float
            Initial time
        stop_condition : function
            A function ``stop_condition(t,y)`` that flips from
            positive to negative values when a certain
            condition is reached. For instance, when
            an element reaches a certain threshold value:

            .. code:: python

                stop_condition = lambda t, y: 0.3 - y[2]

        return_compartments : list, default = None
            A list of compartments for which the result should be returned.
            If ``return_compartments`` is None, all compartments will
            be returned.
        adopt_final_state : bool, default = False
            Whether or not to adopt the final state of the integration
        solve_ivp_kwargs : dict
            Keyword arguments that will be passed to
            ``scipy.integrate.solve_ivp``.

        Returns
        =======
        t : float
            Times when the system state was recorded
            The final time corresponds to the time when
            the stop condition was reached.
        y : numpy.ndarray
            First axis accesses system states, second
            axis refers to the system state corresponding
            to the time points in `t`.
        """

        dydt = self.get_numerical_dydt()
        self.t0 = t0

        time_points, result = integrate_ivp_until(dydt, self.t0, self.y0, stop_condition, solve_ivp_kwargs)

        if adopt_final_state:
            self.t0 = time_points[-1]
            self.y0 = result[:,-1]

        if return_compartments is not None:
            ndx = [self.get_compartment_id(C) for C in return_compartments]
            result = result[ndx,:]

        return time_points, result

    def integrate_until(self,
                        t0,
                        stop_condition,
                        return_compartments=None,
                        adopt_final_state=False,
                        solve_ivp_kwargs={},
                        ):
        r"""
        Integrate the system until a stop condition is reached,
        using ``scipy.integrate.solve_ivp``.

        Parameters
        ==========
        t0 : float
            Initial time
        stop_condition : function
            A function ``stop_condition(t,y)`` that flips from
            positive to negative values when a certain
            condition is reached. For instance, when
            an element reaches a certain threshold value:

            .. code:: python

                stop_condition = lambda t, y: 0.3 - y[2]

        return_compartments : list, default = None
            A list of compartments for which the result should be returned.
            If ``return_compartments`` is None, all compartments will
            be returned.
        adopt_final_state : bool, default = False
            Whether or not to adopt the final state of the integration
        solve_ivp_kwargs : dict
            Keyword arguments that will be passed to
            ``scipy.integrate.solve_ivp``.

        Returns
        =======
        t : float
            Times when the system state was recorded
            The final time corresponds to the time when
            the stop condition was reached.
        result : dict
            Dictionary mapping compartments to time series.
        """
        if return_compartments is None:
            return_compartments = self.compartments

        t, result = self.integrate_and_return_by_index_until(
                            t0,
                            stop_condition,
                            return_compartments,
                            adopt_final_state,
                            solve_ivp_kwargs,
                        )

        result_dict = {}
        for icomp, compartment in enumerate(return_compartments):
            result_dict[compartment] = result[icomp,:]

        return t, result_dict

def time_leap_newton(t0, y0, get_event_rates, rand=None):
    """
    Compute a time leap for time-varying rate functions
    based on Gillespie's SSA using Newton's method
    on a quad-integrator.

    Parameters
    ==========
    t0 : float
        The current time
    y0 : numpy.ndarray
        Current state of the system
    get_event_rates : function
        A function that takes time t and state y
        as input and returns an array of corresponding
        event rates
    rand : float, default = None
        A random number from the unit interval.

    Returns
    =======
    new_t : float
        A new time.
    """
    integrand = lambda _t : get_event_rates(_t, y0).sum()
    integral = lambda _t : quad(integrand, t0, _t)[0]

    if rand is None:
        rand = np.random.rand()
    else:
        assert(rand >= 0)
        assert(rand < 1)
    _1_minus_r = 1 - rand

    rootfunction = lambda _t: - np.log(_1_minus_r) - integral(_t)
    new_t = newton(rootfunction, t0, fprime=lambda _t: -integrand(_t))

    return new_t

def time_leap_ivp(t0, y0, get_event_rates, rand=None):
    """
    Compute a time leap for time-varying rate functions
    based on Gillespie's SSA using Newton's method
    on a quad-integrator.

    Parameters
    ==========
    t0 : float
        The current time
    y0 : numpy.ndarray
        Current state of the system
    get_event_rates : function
        A function that takes time t and state y
        as input and returns an array of corresponding
        event rates
    rand : float, default = None
        A random number from the unit interval.

    Returns
    =======
    new_t : float
        A new time.
    """
    integrand = lambda _t, _y: [get_event_rates(_t, y0).sum()]
    initial = integrand(t0,None)

    if rand is None:
        rand = np.random.rand()
    else:
        assert(rand >= 0)
        assert(rand < 1)
    _1_minus_r = 1 - rand

    rootfunction = lambda _t, _y: - np.log(_1_minus_r) - _y[0]
    rootfunction.terminal = True
    result = solve_ivp(integrand,[t0,np.inf],y0=[0],method='RK23',events=rootfunction)

    return result.t_events[0][0]

if __name__ == "__main__":     # pragma: no cover

    from epipack import SISModel

    SIS = SISModel(infection_rate=2.0,recovery_rate=1)
    SIS.set_initial_conditions({
                    'S': 0.99,
                    'I': 0.01,
                })


    tt = np.linspace(0,10,100)
    result = integrate_dopri5(SIS.dydt, tt, SIS.y0)

    import matplotlib.pyplot as pl

    pl.plot(tt, result[0,:])
    pl.plot(tt, result[1,:])

    pl.show()



# -*- coding: utf-8 -*-
"""
Build, simulate, and analyze epidemiological models.
"""

from .metadata import *

from .numeric_matrix_epi_models import (
        MatrixEpiModel,
        MatrixSIModel,
        MatrixSISModel,
        MatrixSIRModel,
        MatrixSIRSModel,
        MatrixSEIRModel,
    )

from .numeric_epi_models import (
        EpiModel,
        SIModel,
        SISModel,
        SIRModel,
        SIRSModel,
        SEIRModel,
    )

from .stochastic_epi_models import (
        StochasticEpiModel,
        StochasticSIModel,
        StochasticSIRModel,
        StochasticSISModel,
    )

from .symbolic_epi_models import (
        SymbolicEpiModel,
        SymbolicODEModel,
        SymbolicSIModel,
        SymbolicSISModel,
        SymbolicSIRModel,
        SymbolicSIRSModel,
        get_temporal_interpolation,
    )

from .symbolic_matrix_epi_models import (
        SymbolicMatrixEpiModel,
        SymbolicMatrixSISModel,
        SymbolicMatrixSIRModel,
        SymbolicMatrixSIRSModel,
        SymbolicMatrixSIModel,
    )

from .networks import (
        get_2D_lattice_links,
        get_random_layout,
        get_grid_layout,
    )




def processes_to_rates(process_list, compartments, ignore_rate_position_checks=False):
    """
    Converts a list of reaction process tuples to rate tuples

    Parameters
    ----------
    process_list : :obj:`list` of :obj:`tuple`
        A list containing reaction processes in terms of tuples.

        .. code:: python

            [
                # transition process
                ( source_compartment, rate, target_compartment),

                # transmission process
                ( coupling_compartment_0, coupling_compartment_1, rate, target_compartment_0, target_ccompartment_1),

                # fission process
                ( source_compartment, rate, target_compartment_0, target_ccompartment_1),

                # fusion process
                ( source_compartment_0, source_compartment_1, rate, target_compartment),

                # death process
                ( source_compartment, rate, None),

                # birth process
                ( None, rate, target_compartment),
            ]

    compartments : :obj:`list` of hashable type
        The compartments of these reaction equations.
    ignore_rate_position_checks : bool, default = False
        This function usually checks whether the rate of
        a reaction is positioned correctly. You can
        turn this behavior off for transition, birth, death, and
        transmission processes. (Useful if you want to define
        symbolic transmission processes that are compartment-dependent).

    Returns
    -------
    quadratic_rates : :obj:`list` of :obj:`tuple`
        Rate tuples for quadratic terms
    linear_rates : :obj:`list` obj:`tuple`
        Rate tuples for linear terms
    """

    quadratic_rates = []
    linear_rates = []

    for process in process_list:

        if len(process) == 3:
            # it's either a transition process or a birth process:
            if ignore_rate_position_checks or process[1] not in compartments:
                linear_rates.extend(transition_processes_to_rates([process]))
            else:
                raise TypeError("Process " + str(tuple(process)) + " is not understood.")

        elif len(process) == 4:
            # it's either a fission process or a fusion process

            if process[1] not in compartments:
                # it's a fission process
                linear_rates.extend(fission_processes_to_rates([process]))
            elif process[2] not in compartments:
                # it's a fusion process
                quadratic_rates.extend(fusion_processes_to_rates([process]))
            else:
                raise TypeError("Process " + str(tuple(process)) + " is not understood because it's ambiguous in whether it's supposed to be a fission or a fusion process. Please use model.add_fission_processes() or model.add_fusion_processes() to set this process.")

        elif len(process) == 5:

            # it's a transmission process
            if ignore_rate_position_checks or process[2] not in compartments:
                quadratic_rates.extend(transmission_processes_to_rates([process]))
            else:
                raise TypeError("Process " + str(tuple(process)) + " is not understood.")

        else:
            raise TypeError("Process " + str(tuple(process)) + " is not understood.")

    return quadratic_rates, linear_rates


def transition_processes_to_rates(process_list):
    """
    Define the transition processes between compartments, including birth and deaths processes.

    Parameters
    ==========

    process_list : :obj:`list` of :obj:`tuple`
        A list of tuples that contains transitions rates in the following format:

        .. code:: python

            [
                ( source_compartment, rate, target_compartment ),
                ...
            ]

    Example
    -------

    For an SEIR model.

    .. code:: python

        transition_processes_to_rates([
            ("E", symptomatic_rate, "I" ),
            ("I", recovery_rate, "R" ),
        ])

    """
    linear_rates = []

    for source, rate, target in process_list:
        if source is None and target is None:
            raise ValueError("The reaction" + str((source, rate, target)) + " is meaningless because there are no reactants.")
        elif source == target:
            raise ValueError("Process "+\
                             str((source, rate, target)) +\
                             " leaves system unchanged")

        elif source is None and target is not None:
            #birth process
            linear_rates.append( (None, target, rate) )
        elif source is not None and target is None:
            #death process
            linear_rates.append( (source, source, -rate) )
        else:
            # source compartment loses an entity
            # target compartment gains one
            linear_rates.append((source, source, -rate))
            linear_rates.append((source, target, +rate))

    return linear_rates


def fission_processes_to_rates(process_list):
    """
    Define linear fission processes between compartments.

    Parameters
    ==========

    process_list : :obj:`list` of :obj:`tuple`
        A list of tuples that contains fission rates in the following format:

        .. code:: python

            [
                (source_compartment, rate, target_compartment_0, target_compartment_1 ),
                ...
            ]

    Example
    -------

    For pure exponential growth of compartment `B`.

    .. code:: python

        epi.set_fission_processes([
            ("B", growth_rate, "B", "B" ),
        ])

    """

    linear_rates = []

    for source, rate, target0, target1 in process_list:

        _s = source
        _t0 = target0
        _t1 = target1

        # source compartment loses an entity
        # target compartments gains one each
        linear_rates.append((_s, _s, -rate))
        linear_rates.append((_s, _t0, +rate))
        linear_rates.append((_s, _t1, +rate))

    return linear_rates

def fusion_processes_to_rates(process_list):
    """
    Define fusion processes between compartments.

    Parameters
    ==========

    process_list : :obj:`list` of :obj:`tuple`
        A list of tuples that contains fission rates in the following format:

        .. code:: python

            [
                (coupling_compartment_0, coupling_compartment_1, rate, target_compartment_0 ),
                ...
            ]

    Example
    -------

    Fusion of reactants "A", and "B" to form "C".

    .. code:: python

        fusion_processes_to_rates([
            ("A", "B", reaction_rate, "C" ),
        ])

    """

    quad_rates = []

    for source0, source1, rate, target in process_list:

        # target compartment gains one entity
        quad_rates.append((source0, source1, target, rate))
        # source compartments lose one entity each
        quad_rates.append((source0, source1, source0, -rate))
        quad_rates.append((source0, source1, source1, -rate))

    return quad_rates

def transmission_processes_to_rates(process_list):
    r"""
    A wrapper to define quadratic process rates through transmission reaction equations.
    Note that in stochastic network/agent simulations, the transmission
    rate is equal to a rate per link. For the mean-field ODEs,
    the rates provided to this function will just be equal 
    to the prefactor of the respective quadratic terms.

    For instance, if you analyze an SIR system and simulate on a network of mean degree :math:`k_0`,
    a basic reproduction number :math:`R_0`, and a recovery rate :math:`\mu`,
    you would define the single link transmission process as 

        .. code:: python

            ("I", "S", R_0/k_0 * mu, "I", "I")

    For the mean-field system here, the corresponding reaction equation would read

        .. code:: python

            ("I", "S", R_0 * mu, "I", "I")

    Parameters
    ----------
    process_list : :obj:`list` of :obj:`tuple`
        A list of tuples that contains transitions rates in the following format:

        .. code:: python

            [
                (source_compartment,
                 target_compartment_initial,
                 rate
                 source_compartment,
                 target_compartment_final,
                 ),
                ...
            ]

    Example
    -------

    For an SEIR model.

    .. code:: python

        transmission_processes_to_rates([
            ("I", "S", +1, "I", "E" ),
        ])

    """

    rate_list = []

    # iterate through processes
    for coupling0, coupling1, rate, affected0, affected1 in process_list:

        _s0 = coupling0
        _s1 = coupling1
        _t0 = affected0
        _t1 = affected1

        reactants = [_s0, _s1]
        products = [_t0, _t1]
        constant = (set.intersection(set(reactants), set(products)))
        if len(constant) == 2 or tuple(reactants) == tuple(products):
            raise ValueError("Process "+\
                             str((coupling0, coupling1, rate, affected0, affected1)) +\
                             " leaves system unchanged")
        elif len(constant) == 1:
            constant = next(iter(constant))
            reactants.remove(constant)
            products.remove(constant)
            _s0 = constant
            _s1 = reactants[0]
            _t1 = products[0]

            # if one compartment (_s0) remains constant,
            # there's no changes in counts of this compartment.
            # Rather the source compartment of the other reactant
            # loses one entity, while the target compartment
            # of this reactant gains one.
            rate_list.append( (_s0, _s1, _s1, -rate) )
            rate_list.append( (_s0, _s1, _t1, +rate) )
        else:

            # when no reactant remains constant,
            # all source compartments lose one entity
            # and all target compartments gain one
            rate_list.append( (_s0, _s1, _s1, -rate) )
            rate_list.append( (_s0, _s1, _t1, +rate) )
            rate_list.append( (_s0, _s1, _s0, -rate) )
            rate_list.append( (_s0, _s1, _t0, +rate) )


    return rate_list

def processes_to_events(process_list, compartments, ignore_rate_position_checks=False):
    """
    Converts a list of reaction process tuples to event tuples

    Parameters
    ----------
    process_list : :obj:`list` of :obj:`tuple`
        A list containing reaction processes in terms of tuples.

        .. code:: python

            [
                # transition process
                ( source_compartment, rate, target_compartment),

                # transmission process
                ( coupling_compartment_0, coupling_compartment_1, rate, target_compartment_0, target_ccompartment_1),

                # fission process
                ( source_compartment, rate, target_compartment_0, target_compartment_1),
                
                # fusion process
                ( source_compartment_0, source_compartment_1, rate, target_compartment),

                # death process
                ( source_compartment, rate, None),

                # birth process
                ( None, rate, target_compartment),
            ]

    compartments : :obj:`list` of hashable type
        The compartments of these reaction equations. 
    ignore_rate_position_checks : bool, default = False
        This function usually checks whether the rate of 
        a reaction is positioned correctly. You can
        turn this behavior off for transition, birth, death, and
        transmission processes. (Useful if you want to define
        symbolic transmission processes that are compartment-dependent).

    Returns
    -------
    quadratic_events : :obj:`list` of :obj:`tuple`
        event tuples for quadratic terms
    linear_events : :obj:`list` obj:`tuple`
        event tuples for linear terms
    """

    quadratic_events = []
    linear_events = []

    for process in process_list:

        if len(process) == 3:
            # it's either a transition process or a birth process:
            if ignore_rate_position_checks or process[1] not in compartments:
                linear_events.extend(transition_processes_to_events([process]))
            else:
                raise TypeError("Process " + str(tuple(process)) + " is not understood.")

        elif len(process) == 4:
            # it's either a fission process or a fusion process

            if process[1] not in compartments:
                # it's a fission process
                linear_events.extend(fission_processes_to_events([process]))
            elif process[2] not in compartments:
                # it's a fusion process
                quadratic_events.extend(fusion_processes_to_events([process]))
            else:
                raise TypeError("Process " + str(tuple(process)) + " is not understood because it's ambiguous in whether it's supposed to be a fission or a fusion process. Please use model.add_fission_processes() or model.add_fusion_processes() to set this process.")

        elif len(process) == 5:

            # it's a transmission process
            if ignore_rate_position_checks or process[2] not in compartments:
                quadratic_events.extend(transmission_processes_to_events([process]))
            else:
                raise TypeError("Process " + str(tuple(process)) + " is not understood.")

        else:
            raise TypeError("Process " + str(tuple(process)) + " is not understood.")

    return quadratic_events, linear_events


def transition_processes_to_events(process_list):
    """
    Define the transition processes between compartments, including birth and deaths processes.

    Parameters
    ==========

    process_list : :obj:`list` of :obj:`tuple`
        A list of tuples that contains transitions rates in the following format:

        .. code:: python

            [
                ( source_compartment, rate, target_compartment ),
                ...
            ]

    Example
    -------

    For an SEIR model.

    .. code:: python

        transition_processes_to_events([
            ("E", symptomatic_rate, "I" ),
            ("I", recovery_rate, "R" ),
        ])

    """
    linear_events = []

    for source, rate, target in process_list:
        if source is None and target is None:
            raise ValueError("The reaction" + str((source, rate, target)) + " is meaningless because there are no reactants.")
        elif source == target:
            raise ValueError("Process "+\
                             str((source, rate, target)) +\
                             " leaves system unchanged")

        elif source is None and target is not None:
            #birth process
            linear_events.append( ( (None,), rate, [(target, +1)] ))
        elif source is not None and target is None:
            #death process
            linear_events.append(( (source,), rate, [(source, -1)] ))
        else:
            # source compartment loses an entity
            # target compartment gains one
            linear_events.append(( (source,), rate, [(source, -1), (target, +1)] ))

    return linear_events


def fission_processes_to_events(process_list):
    """
    Define linear fission processes between compartments.

    Parameters
    ==========

    process_list : :obj:`list` of :obj:`tuple`
        A list of tuples that contains fission rates in the following format:

        .. code:: python

            [
                (source_compartment, rate, target_compartment_0, target_compartment_1 ),
                ...
            ]

    Example
    -------

    For pure exponential growth of compartment `B`.

    .. code:: python

        epi.set_fission_processes([
            ("B", growth_rate, "B", "B" ),
        ])

    """

    linear_events = []
    
    for source, rate, target0, target1 in process_list:

        _s = source
        _t0 = target0
        _t1 = target1
        
        # source compartment loses an entity
        # target compartments gains one each
        linear_events.append(( (_s,), rate, [(_s, -1), (_t0, +1), (_t1, +1)] ))

    return linear_events

def fusion_processes_to_events(process_list):
    """
    Define fusion processes between compartments.

    Parameters
    ==========

    process_list : :obj:`list` of :obj:`tuple`
        A list of tuples that contains fission rates in the following format:

        .. code:: python

            [
                (coupling_compartment_0, coupling_compartment_1, rate, target_compartment_0 ),
                ...
            ]

    Example
    -------

    Fusion of reactants "A", and "B" to form "C".

    .. code:: python

        fusion_processes_to_events([
            ("A", "B", reaction_rate, "C" ),
        ])

    """

    quad_events = []
    
    for source0, source1, rate, target in process_list:
        
        # target compartment gains one entity
        # source compartments lose one entity each
        quad_events.append(( (source0, source1), rate, [(target, +1),(source0, -1), (source1, -1)] ))

    return quad_events

def transmission_processes_to_events(process_list):
    r"""
    A wrapper to define quadratic process events through transmission reaction equations.
    Note that in stochastic network/agent simulations, the transmission
    rate is equal to a rate per link. For the mean-field ODEs,
    the rates provided to this function will just be equal 
    to the prefactor of the respective quadratic terms.

    For instance, if you analyze an SIR system and simulate on a network of mean degree :math:`k_0`,
    a basic reproduction number :math:`R_0`, and a recovery rate :math:`\mu`,
    you would define the single link transmission process as 

        .. code:: python

            ("I", "S", R_0/k_0 * mu, "I", "I")

    For the mean-field system here, the corresponding reaction equation would read
        
        .. code:: python
            
            ("I", "S", R_0 * mu, "I", "I")

    Parameters
    ----------
    process_list : :obj:`list` of :obj:`tuple`
        A list of tuples that contains transitions events in the following format:

        .. code:: python

            [
                (source_compartment, 
                 target_compartment_initial,
                 rate 
                 source_compartment, 
                 target_compartment_final, 
                 ),
                ...
            ]

    Example
    -------

    For an SEIR model.

    .. code:: python

        transmission_processes_to_events([
            ("I", "S", +1, "I", "E" ),
        ])

    """

    event_list = []

    # iteevent through processes
    for coupling0, coupling1, rate, affected0, affected1 in process_list:

        _s0 = coupling0
        _s1 = coupling1
        _t0 = affected0
        _t1 = affected1

        reactants = [_s0, _s1]
        products = [_t0, _t1]
        constant = (set.intersection(set(reactants), set(products)))        
        if len(constant) == 2 or tuple(reactants) == tuple(products):
            raise ValueError("Process "+\
                             str((coupling0, coupling1, rate, affected0, affected1)) +\
                             " leaves system unchanged")
        elif len(constant) == 1:
            constant = next(iter(constant))
            reactants.remove(constant)
            products.remove(constant)
            _s0 = constant
            _s1 = reactants[0]
            _t1 = products[0]

            # if one compartment (_s0) remains constant,
            # there's no changes in counts of this compartment.
            # Rather the source compartment of the other reactant
            # loses one entity, while the target compartment
            # of this reactant gains one.
            event_list.append(( (_s0, _s1), rate, [(_s1, -1),(_t1, +1)]  ))
        else:

            # when no reactant remains constant,
            # all source compartments lose one entity
            # and all target compartments gain one
            event_list.append(( (_s0, _s1), rate, [(_s1, -1), (_t1, +1), (_s0, -1), (_t0, +1)] ))


    return event_list



import unittest

import numpy as np

from epipack.vis import (
            visualize_reaction_diffusion,
            visualize,
            get_grid_layout,
            get_random_layout,
        )

from epipack import StochasticSIRModel, StochasticEpiModel, get_2D_lattice_links
import pyglet
import gc

class VisTest(unittest.TestCase):

    def test_grid_well_mixed(self):
        N = 20 * 20
        network = get_grid_layout(N)

        R0 = 3
        recovery_rate = 1/8
        model = StochasticSIRModel(N,R0,recovery_rate)
        model.set_random_initial_conditions({'I':20,'S':N-20})

        sampling_dt = 2

        visualize(model,network,sampling_dt,
                  ignore_plot_compartments=['S'],
                  config={
                        'draw_nodes_as_rectangles':True,
                        'show_legend':False,
                     }
                  )

    def test_reaction_diffustion(self):
        pass

    def test_lattice(self):
        # define links and network layout
        N_side = 20
        N = N_side**2
        links = get_2D_lattice_links(N_side, periodic=True, diagonal_links=True)
        network = get_grid_layout(N)

        # define model
        R0 = 3; recovery_rate = 1/8
        model = StochasticSIRModel(N,R0,recovery_rate,
                                   edge_weight_tuples=links)
        model.set_random_initial_conditions({'I':20,'S':N-20})

        sampling_dt = 2

        visualize(model,network,sampling_dt,
                    config={
                     'draw_nodes_as_rectangles':True,
                     'draw_links':False,
                     'show_legend':False,
                     'show_curves':False,
                   }
              )

    def test_network(self):
        N = 1000
        k = 4
        links = []
        for i in range(N):
            neighs = np.random.randint(0,N-1,size=(k,),dtype=int)
            neighs[neighs>=i] += 1
            for neigh in neighs:
                links.append((i,int(neigh),1.0))

        network = get_random_layout(N,links,windowwidth=500)

        model = StochasticEpiModel(list("SIRXTQ"),
                                   N=len(network['nodes']),
                                   directed=True,
                                   edge_weight_tuples=links,
                                   )
        k0 = model.out_degree.mean()
        R0 = 10
        recovery_rate = 1/8
        quarantine_rate = 1/16
        tracing_rate = 1/2
        waning_immunity_rate = 1/14
        infection_rate = R0 * (recovery_rate) / k0
        model.set_node_transition_processes([
                ("I",recovery_rate,"R"),
                ("I",quarantine_rate,"T"),
                ("T",tracing_rate,"X"),
                ("Q",waning_immunity_rate,"S"),
                ("X",recovery_rate,"R"),
                ])
        model.set_link_transmission_processes([("I","S",infection_rate,"I","I")])
        model.set_conditional_link_transmission_processes({
            ("T", "->", "X") : [
                     ("X","I",0.5,"X","T"),
                     #("X","S",0.5,"X","Q"),
                     ],
            })
        model.set_random_initial_conditions({'I':20,'S':N-20})

        sampling_dt = 0.08

        visualize(model,network,sampling_dt,ignore_plot_compartments=['S'],quarantine_compartments=['X', 'T', 'Q'])


if __name__ == "__main__":

    T = VisTest()
    T.test_network()
    T.test_lattice()
    T.test_grid_well_mixed()
